from typing import Optional
from rl4co.models.common.constructive.base import ConstructivePolicy

from .decoder import NonAutoregressiveDecoder
from .encoder import NonAutoregressiveEncoder


class NonAutoregressivePolicy(ConstructivePolicy):
    """Template class for an nonautoregressive policy, simple wrapper around
    :class:`rl4co.models.common.constructive.base.ConstructivePolicy`.
    """

    def __init__(
        self,
        encoder: NonAutoregressiveEncoder,
        decoder: Optional[NonAutoregressiveDecoder] = None,
        env_name: str = "tsp",
        temperature: float = 1.0,
        tanh_clipping: float = 0,
        mask_logits: bool = True,
        train_decode_type: str = "sampling",
        val_decode_type: str = "greedy",
        test_decode_type: str = "greedy",
        **unused_kw,
    ):
        # If decoder is not passed, we default to the non-autoregressive decoder that decodes the heatmap
        if decoder is None:
            decoder = NonAutoregressiveDecoder()

        super(NonAutoregressivePolicy, self).__init__(
            encoder=encoder,
            decoder=decoder,
            env_name=env_name,
            temperature=temperature,
            tanh_clipping=tanh_clipping,
            mask_logits=mask_logits,
            train_decode_type=train_decode_type,
            val_decode_type=val_decode_type,
            test_decode_type=test_decode_type,
            **unused_kw,
        )



from typing import Callable, Union

import torch

from tensordict.tensordict import TensorDict
from torch.distributions import Uniform

from rl4co.envs.common.utils import Generator, get_sampler
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class MDCPDPGenerator(Generator):
    """Data generator for the Multi Depot Capacitated Pickup and Delivery Problem (MDCPDP) environment.

    Args:
        num_loc: number of locations (customers)
        min_loc: minimum value for the location coordinates
        max_loc: maximum value for the location coordinates, default is 150 insted of 1.0, will be scaled
        loc_distribution: distribution for the location coordinates
        num_agents: number of depots, each depot has one vehicle
        depot_mode: mode for the depot, either single or multiple
        depod_distribution: distribution for the depot coordinates
        min_capacity: minimum value of the capacity
        max_capacity: maximum value of the capacity
        min_lateness_weight: minimum value of the lateness weight
        max_lateness_weight: maximum value of the lateness weight
        latebess_weight_distribution: distribution for the lateness weight

    Returns:
        A TensorDict with the following keys:
            locs [batch_size, num_loc, 2]: locations of each customer
            depot [batch_size, num_agents, 2]: locations of each depot
            capacity [batch_size, 1]: capacity of the vehicle
            lateness_weight [batch_size, 1]: weight of the lateness cost
    """

    def __init__(
        self,
        num_loc: int = 20,
        min_loc: float = 0.0,
        max_loc: float = 1.0,
        loc_distribution: Union[int, float, str, type, Callable] = Uniform,
        num_agents: int = 5,
        depot_mode: str = "multiple",
        depot_distribution: Union[int, float, str, type, Callable] = Uniform,
        min_capacity: int = 3,
        max_capacity: int = 3,
        min_lateness_weight: float = 1.0,
        max_lateness_weight: float = 1.0,
        lateness_weight_distribution: Union[int, float, str, type, Callable] = Uniform,
        **kwargs,
    ):
        self.num_loc = num_loc
        self.min_loc = min_loc
        self.max_loc = max_loc
        self.depot_mode = depot_mode
        self.num_agents = num_agents
        self.min_capacity = min_capacity
        self.max_capacity = max_capacity
        self.min_lateness_weight = min_lateness_weight
        self.max_lateness_weight = max_lateness_weight

        # Number of locations must be even
        if num_loc % 2 != 0:
            log.warn(
                "Number of locations must be even. Adding 1 to the number of locations."
            )
            self.num_loc += 1

        # Check depot mode validity
        assert depot_mode in ["single", "multiple"], f"Invalid depot mode: {depot_mode}"

        # Location distribution
        if kwargs.get("loc_sampler", None) is not None:
            self.loc_sampler = kwargs["loc_sampler"]
        else:
            self.loc_sampler = get_sampler(
                "loc", loc_distribution, min_loc, max_loc, **kwargs
            )

        # Depot distribution
        if kwargs.get("depot_sampler", None) is not None:
            self.depot_sampler = kwargs["depot_sampler"]
        else:
            self.depot_sampler = get_sampler(
                "depot", depot_distribution, min_loc, max_loc, **kwargs
            )

        # Lateness weight distribution
        if kwargs.get("lateness_weight_sampler", None) is not None:
            self.lateness_weight_sampler = kwargs["lateness_weight_sampler"]
        else:
            self.lateness_weight_sampler = get_sampler(
                "lateness_weight",
                lateness_weight_distribution,
                min_lateness_weight,
                max_lateness_weight,
                **kwargs,
            )

    def _generate(self, batch_size) -> TensorDict:
        # Sample locations
        locs = self.loc_sampler.sample((*batch_size, self.num_loc, 2))

        # Sample depot
        if self.depot_mode == "single":
            depot = self.depot_sampler.sample((*batch_size, 2))[:, None, :].repeat(
                1, self.num_agents, 1
            )
        else:
            depot = self.depot_sampler.sample((*batch_size, self.num_agents, 2))

        # Sample capacity
        capacity = torch.randint(
            self.min_capacity, self.max_capacity + 1, size=(*batch_size, self.num_agents)
        )

        # Sample lateness weight
        lateness_weight = self.lateness_weight_sampler.sample((*batch_size, 1))

        return TensorDict(
            {
                "locs": locs,
                "depot": depot,
                "capacity": capacity,
                "lateness_weight": lateness_weight,
            },
            batch_size=batch_size,
        )



import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class MultiHeadAttentionMDAM(nn.Module):
    def __init__(self, embed_dim, n_heads, last_one=False, sdpa_fn=None):
        super(MultiHeadAttentionMDAM, self).__init__()

        if sdpa_fn is not None:
            log.warning("sdpa_fn is not used in this implementation")

        self.embed_dim = embed_dim
        self.n_heads = n_heads

        self.norm_factor = 1 / math.sqrt(embed_dim)  # See Attention is all you need

        self.W_query = nn.Parameter(torch.Tensor(n_heads, embed_dim, embed_dim))
        self.W_key = nn.Parameter(torch.Tensor(n_heads, embed_dim, embed_dim))
        self.W_val = nn.Parameter(torch.Tensor(n_heads, embed_dim, embed_dim))
        self.W_out = nn.Parameter(torch.Tensor(n_heads, embed_dim, embed_dim))

        self.init_parameters()
        self.last_one = last_one

    def init_parameters(self):
        for param in self.parameters():
            stdv = 1.0 / math.sqrt(param.size(-1))
            param.data.uniform_(-stdv, stdv)

    def forward(self, q, h=None, mask=None):
        if h is None:
            h = q  # compute self-attention

        # h should be (batch_size, graph_size, input_dim)
        batch_size, graph_size, input_dim = h.size()
        n_query = q.size(1)
        assert q.size(0) == batch_size
        assert q.size(2) == input_dim
        assert input_dim == self.embed_dim, "Wrong embedding dimension of input"

        hflat = h.contiguous().view(-1, input_dim)
        qflat = q.contiguous().view(-1, input_dim)

        # last dimension can be different for keys and values
        shp = (self.n_heads, batch_size, graph_size, -1)
        shp_q = (self.n_heads, batch_size, n_query, -1)

        # Calculate queries, (n_heads, n_query, graph_size, key/val_size)
        Q = torch.matmul(qflat, self.W_query).view(shp_q)
        # Calculate keys and values (n_heads, batch_size, graph_size, key/val_size)
        K = torch.matmul(hflat, self.W_key).view(shp)
        V = torch.matmul(hflat, self.W_val).view(shp)

        # Calculate compatibility (n_heads, batch_size, n_query, graph_size)
        compatibility = self.norm_factor * torch.matmul(Q, K.transpose(2, 3))

        # Optionally apply mask to prevent attention
        if mask is not None:
            mask = mask.view(1, batch_size, n_query, graph_size).expand_as(compatibility)
            compatibility[mask] = float("-inf")

        attn = F.softmax(compatibility, dim=-1)

        # If there are nodes with no neighbours then softmax returns nan so we fix them to 0
        if mask is not None:
            attnc = attn.clone()
            attnc[mask] = 0
            attn = attnc

        heads = torch.matmul(attn, V)

        out = torch.mm(
            heads.permute(1, 2, 0, 3)
            .contiguous()
            .view(-1, self.n_heads * self.embed_dim),
            self.W_out.view(-1, self.embed_dim),
        ).view(batch_size, n_query, self.embed_dim)
        if self.last_one:
            return (out, attn, V)
        return out



import itertools

from math import factorial
from typing import Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.data.dataset import FastTdDataset
from rl4co.envs.common.base import RL4COEnvBase

from .generator import FFSPGenerator


class FFSPEnv(RL4COEnvBase):
    """Flexible Flow Shop Problem (FFSP) environment.
    The goal is to schedule a set of jobs on a set of machines such that the makespan is minimized.

    Observations:
        - time index
        - sub time index
        - batch index
        - machine index
        - schedule
        - machine wait step
        - job location
        - job wait step
        - job duration

    Constraints:
        - each job has to be processed on each machine in a specific order
        - the machine has to be available to process the job
        - the job has to be available to be processed

    Finish Condition:
        - all jobs are scheduled

    Reward:
        - (minus) the makespan of the schedule

    Args:
        generator: FFSPGenerator instance as the data generator
        generator_params: parameters for the generator
    """

    name = "ffsp"

    def __init__(
        self,
        generator: FFSPGenerator = None,
        generator_params: dict = {},
        **kwargs,
    ):
        super().__init__(check_solution=False, dataset_cls=FastTdDataset, **kwargs)
        if generator is None:
            generator = FFSPGenerator(**generator_params)
        self.generator = generator

        self.num_stage = generator.num_stage
        self.num_machine = generator.num_machine
        self.num_job = generator.num_job
        self.num_machine_total = generator.num_machine_total
        self.tables = None
        self.step_cnt = None
        self.flatten_stages = generator.flatten_stages

        self._make_spec(generator)

    def get_num_starts(self, td):
        return factorial(self.num_machine)

    def select_start_nodes(self, td, num_starts):
        self.tables.augment_machine_tables(num_starts)
        selected = torch.full((num_starts * td.size(0),), self.num_job)
        return selected

    def _move_to_next_machine(self, td):
        batch_size = td.batch_size
        batch_idx = torch.arange(*batch_size, dtype=torch.long, device=td.device)

        time_idx = td["time_idx"]
        machine_idx = td["machine_idx"]
        sub_time_idx = td["sub_time_idx"]

        machine_wait_step = td["machine_wait_step"]
        job_wait_step = td["job_wait_step"]
        job_location = td["job_location"]

        ready = torch.flatten(td["done"])
        idx = torch.flatten(batch_idx)
        # select minibatch instances that need updates (are not done)
        idx = idx[~ready]

        while ~ready.all():
            # increment the stage-machine counter
            new_sub_time_idx = sub_time_idx[idx] + 1
            # increment time if all machines-stage combinations have been candidates
            step_time_required = new_sub_time_idx == self.num_machine_total
            time_idx[idx] += step_time_required.long()
            # in this case set the machine-stage counter to zero again
            new_sub_time_idx[step_time_required] = 0
            # update machine-stage counter
            sub_time_idx[idx] = new_sub_time_idx
            # determine current machine candidate
            new_machine_idx = self.tables.get_machine_index(idx, new_sub_time_idx)
            machine_idx[idx] = new_machine_idx

            # decrease machine wait time by 1 if instance transitioned to new time step
            machine_wait_steps = machine_wait_step[idx, :]
            machine_wait_steps[step_time_required, :] -= 1
            machine_wait_steps[machine_wait_steps < 0] = 0
            machine_wait_step[idx, :] = machine_wait_steps

            # decrease job wait time by 1 if instance transitioned to new time step
            job_wait_steps = job_wait_step[idx, :]
            job_wait_steps[step_time_required, :] -= 1
            job_wait_steps[job_wait_steps < 0] = 0
            job_wait_step[idx, :] = job_wait_steps
            # machine is ready if its wait time is zero
            machine_ready = machine_wait_step[idx, new_machine_idx] == 0
            # job is ready if the current stage matches the stage of the job and
            # its wait time is zero (no operation of previous stage is in process)
            new_stage_idx = self.tables.get_stage_index(new_sub_time_idx)
            job_ready_1 = job_location[idx, : self.num_job] == new_stage_idx[:, None]
            job_ready_2 = job_wait_step[idx, : self.num_job] == 0
            job_ready = (job_ready_1 & job_ready_2).any(dim=-1)
            # instance ready if at least one job and the current machine are ready
            ready = machine_ready & job_ready
            assert ready.shape == idx.shape
            idx = idx[~ready]

        return td.update(
            {
                "time_idx": time_idx,
                "sub_time_idx": sub_time_idx,
                "machine_idx": machine_idx,
                "machine_wait_step": machine_wait_step,
                "job_wait_step": job_wait_step,
            }
        )

    def pre_step(self, td: TensorDict) -> TensorDict:
        batch_size = td.batch_size
        batch_idx = torch.arange(*batch_size, dtype=torch.long, device=td.device)
        sub_time_idx = td["sub_time_idx"]
        # update machine index
        td["machine_idx"] = self.tables.get_machine_index(batch_idx, sub_time_idx)
        # update action mask and stage machine indx
        td = self._update_step_state(td)
        # perform some checks
        assert (td["stage_idx"] == 0).all(), "call pre_step only at beginning of env"
        assert torch.all(td["stage_machine_idx"] == td["machine_idx"])
        # return updated td
        return td

    def _update_step_state(self, td):
        batch_size = td.batch_size
        batch_idx = torch.arange(*batch_size, dtype=torch.long, device=td.device)

        sub_time_idx = td["sub_time_idx"]
        job_location = td["job_location"]
        job_wait_step = td["job_wait_step"]
        if len(td["done"].shape) == 2:
            done = td["done"].squeeze(1)
        else:
            done = td["done"]

        # update stage
        stage_idx = self.tables.get_stage_index(sub_time_idx)
        stage_machine_idx = self.tables.get_stage_machine_index(batch_idx, sub_time_idx)

        job_loc = job_location[:, : self.num_job]
        job_wait_time = job_wait_step[:, : self.num_job]
        # determine if job can be scheduled in current stage
        # (i.e. previous stages are completed)
        job_in_stage = job_loc == stage_idx[:, None]
        job_not_waiting = job_wait_time == 0
        # job can be scheduled if in current stage and not waiting
        job_available = job_in_stage & job_not_waiting
        # determine instance for which waiting is allowed. This is the case if either
        # 1.) any of its jobs need to be scheduled in a previous stage,
        # 2.) any of the jobs wait for an operation of the preceeding stage to finish
        # 3.) the instance is done.
        job_in_previous_stages = (job_loc < stage_idx[:, None]).any(dim=-1)
        job_waiting_in_stage = (job_in_stage & (job_wait_time > 0)).any(dim=-1)
        wait_allowed = job_in_previous_stages + job_waiting_in_stage + done

        job_enable = torch.cat((job_available, wait_allowed[:, None]), dim=-1)
        job_mask = torch.full_like(td["action_mask"], 0).masked_fill(job_enable, 1)
        assert torch.logical_or((job_mask[:, :-1].sum(1) > 0), done).all()
        return td.update(
            {
                "action_mask": job_mask,
                "stage_idx": stage_idx,
                "stage_machine_idx": stage_machine_idx,
            }
        )

    def _step(self, td: TensorDict) -> TensorDict:
        self.step_cnt += 1
        batch_size = td.batch_size
        batch_idx = torch.arange(*batch_size, dtype=torch.long, device=td.device)

        # job_idx is the action from the model
        job_idx = td["action"]
        time_idx = td["time_idx"]
        machine_idx = td["machine_idx"]

        # increment the operation counter of the selected job
        td["job_location"][batch_idx, job_idx] += 1
        # td["job_location"][:, :-1].clip_(0, self.num_stage)
        # assert (td["job_location"][:, : self.num_job] <= self.num_stage).all()
        # insert start time of the selected job in the schedule
        td["schedule"][batch_idx, machine_idx, job_idx] = time_idx
        # get the duration of the selected job
        job_length = td["job_duration"][batch_idx, job_idx, machine_idx]
        # set the number of time steps until the selected machine is available again
        td["machine_wait_step"][batch_idx, machine_idx] = job_length

        # set the number of time steps until the next operation of the job can be started
        td["job_wait_step"][batch_idx, job_idx] = job_length
        # determine whether all jobs are scheduled
        td["done"] = (td["job_location"][:, : self.num_job] == self.num_stage).all(dim=-1)

        if td["done"].all():
            pass
        else:
            td = self._move_to_next_machine(td)
            td = self._update_step_state(td)

        if td["done"].all():
            # determine end times of ops by adding the durations to their start times
            end_schedule = td["schedule"] + td["job_duration"].permute(0, 2, 1)
            # exclude dummy job and determine the makespan per job
            end_time_max, _ = end_schedule[:, :, : self.num_job].max(dim=-1)
            # determine the max makespan of all jobs
            end_time_max, _ = end_time_max.max(dim=-1)
            reward = -end_time_max.to(torch.float32)
            td.set("reward", reward)

        return td

    def _reset(
        self, td: Optional[TensorDict] = None, batch_size: Optional[list] = None
    ) -> TensorDict:
        """
        Args:

        Returns:
            - stage_table [batch_size, num_stage * num_machine]
            - machine_table [batch_size, num_machine * num_stage]
            - stage_machine_idx [batch_size, num_stage * num_machine]
            - time_idx [batch_size]
            - sub_time_idx [batch_size]
            - batch_idx [batch_size]
            - machine_idx [batch_size]
            - schedule [batch_size, num_machine_total, num_job+1]
            - machine_wait_step [batch_size, num_machine_total]
            - job_location [batch_size, num_job+1]
            - job_wait_step [batch_size, num_job+1]
            - job_duration [batch_size, num_job+1, num_machine * num_stage]
        """
        device = td.device

        self.step_cnt = 0
        self.tables = IndexTables(self)
        # reset tables to undo the augmentation
        # self.tables._reset(device=self.device)
        self.tables.set_bs(batch_size[0])

        # Init index record tensor
        time_idx = torch.zeros(size=(*batch_size,), dtype=torch.long, device=device)
        sub_time_idx = torch.zeros(size=(*batch_size,), dtype=torch.long, device=device)

        # Scheduling status information
        schedule = torch.full(
            size=(*batch_size, self.num_machine_total, self.num_job + 1),
            dtype=torch.long,
            device=device,
            fill_value=-999999,
        )
        machine_wait_step = torch.zeros(
            size=(*batch_size, self.num_machine_total),
            dtype=torch.long,
            device=device,
        )
        job_location = torch.zeros(
            size=(*batch_size, self.num_job + 1),
            dtype=torch.long,
            device=device,
        )
        job_wait_step = torch.zeros(
            size=(*batch_size, self.num_job + 1),
            dtype=torch.long,
            device=device,
        )
        job_duration = torch.empty(
            size=(*batch_size, self.num_job + 1, self.num_machine * self.num_stage),
            dtype=torch.long,
            device=device,
        )
        job_duration[..., : self.num_job, :] = td["run_time"]
        job_duration[..., self.num_job, :] = 0

        # Finish status information
        reward = torch.full(
            size=(*batch_size,),
            dtype=torch.float32,
            device=device,
            fill_value=float("-inf"),
        )
        done = torch.full(
            size=(*batch_size,),
            dtype=torch.bool,
            device=device,
            fill_value=False,
        )

        action_mask = torch.ones(
            size=(*batch_size, self.num_job + 1), dtype=bool, device=device
        )
        action_mask[..., -1] = 0

        batch_idx = torch.arange(*batch_size, dtype=torch.long, device=td.device)
        stage_idx = self.tables.get_stage_index(sub_time_idx)
        machine_idx = self.tables.get_machine_index(batch_idx, sub_time_idx)
        stage_machine_idx = self.tables.get_stage_machine_index(batch_idx, sub_time_idx)

        return TensorDict(
            {
                # Index information
                "stage_idx": stage_idx,
                "time_idx": time_idx,
                "sub_time_idx": sub_time_idx,
                "machine_idx": machine_idx,
                "stage_machine_idx": stage_machine_idx,
                # Scheduling status information
                "schedule": schedule,
                "machine_wait_step": machine_wait_step,
                "job_location": job_location,
                "job_wait_step": job_wait_step,
                "job_duration": job_duration,
                # Finish status information
                "reward": reward,
                "done": done,
                "action_mask": action_mask,
            },
            batch_size=batch_size,
        )

    def _make_spec(self, generator: FFSPGenerator):
        self.observation_spec = Composite(
            time_idx=Unbounded(
                shape=(1,),
                dtype=torch.int64,
            ),
            sub_time_idx=Unbounded(
                shape=(1,),
                dtype=torch.int64,
            ),
            batch_idx=Unbounded(
                shape=(1,),
                dtype=torch.int64,
            ),
            machine_idx=Unbounded(
                shape=(1,),
                dtype=torch.int64,
            ),
            schedule=Unbounded(
                shape=(generator.num_machine_total, generator.num_job + 1),
                dtype=torch.int64,
            ),
            machine_wait_step=Unbounded(
                shape=(generator.num_machine_total),
                dtype=torch.int64,
            ),
            job_location=Unbounded(
                shape=(generator.num_job + 1),
                dtype=torch.int64,
            ),
            job_wait_step=Unbounded(
                shape=(generator.num_job + 1),
                dtype=torch.int64,
            ),
            job_duration=Unbounded(
                shape=(
                    generator.num_job + 1,
                    generator.num_machine * generator.num_stage,
                ),
                dtype=torch.int64,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(1,),
            dtype=torch.int64,
            low=0,
            high=generator.num_machine_total,
        )
        self.reward_spec = Unbounded(shape=(1,))
        self.done_spec = Unbounded(shape=(1,), dtype=torch.bool)

    def _get_reward(self, td, actions) -> TensorDict:
        return td["reward"]


class IndexTables:
    def __init__(self, env: FFSPEnv):
        self.stage_table = torch.arange(
            env.num_stage, dtype=torch.long, device=env.device
        ).repeat_interleave(env.num_machine)

        # determine the increment of machine ids between stages, i.e. [0,4,8]
        # for instances with 4 machines and three stages
        start_sub_ids = torch.tensor(
            list(range(0, env.num_machine * env.num_stage, env.num_machine)),
            dtype=torch.long,
            device=env.device,
        ).repeat_interleave(env.num_machine)
        # generate all possible permutations of the machine ids and add the stage increment to it
        # (num_permutations, total_machines)
        permutations = torch.tensor(
            list(itertools.permutations(list(range(env.num_machine)))),
            dtype=torch.long,
            device=env.device,
        ).repeat(1, env.num_stage)
        self.machine_table = permutations + start_sub_ids[None]

        if env.flatten_stages:
            # when flatting stages, every machine in each stage is treated as a distinct entity (no shared embeddings)
            # Therefore, all machine need a unique index which is the same as the machine table
            self.stage_machine_table = self.machine_table
        else:
            # when we do not flatten the stages, machines of different stages with the same subtime index
            # share an embedding. In this case, they need the same index (i.e. leave out the stage increment)
            self.stage_machine_table = permutations

    def set_bs(self, bs):
        self.bs = bs

    def get_stage_index(self, sub_time_idx):
        return self.stage_table[sub_time_idx]

    def get_machine_index(self, idx, sub_time_idx):
        pomo_idx = idx // self.bs

        return self.machine_table[pomo_idx, sub_time_idx]

    def get_stage_machine_index(self, idx, sub_time_idx):
        pomo_idx = idx // self.bs
        return self.stage_machine_table[pomo_idx, sub_time_idx]



import math

import torch
import torch.nn as nn

from tensordict import TensorDict
from torch import Tensor

from rl4co.models.common.improvement.base import ImprovementDecoder
from rl4co.models.nn.attention import MultiHeadCompat
from rl4co.models.nn.mlp import MLP
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class DACTDecoder(ImprovementDecoder):
    """
    DACT decoder based on Ma et al. (2021)
    Given the environment state and the dual sets of embeddings (PFE, NFE embeddings), compute the logits for
    selecting two nodes for the 2-opt local search from the current solution


    Args:
        embed_dim: Embedding dimension
        num_heads: Number of attention heads
    """

    def __init__(
        self,
        embed_dim: int = 64,
        num_heads: int = 4,
    ):
        super().__init__()
        self.embed_dim = embed_dim
        self.n_heads = num_heads
        self.hidden_dim = embed_dim

        # for MHC sublayer (NFE aspect)
        self.compater_node = MultiHeadCompat(
            num_heads, embed_dim, embed_dim, embed_dim, embed_dim
        )

        # for MHC sublayer (PFE aspect)
        self.compater_pos = MultiHeadCompat(
            num_heads, embed_dim, embed_dim, embed_dim, embed_dim
        )

        self.norm_factor = 1 / math.sqrt(1 * self.hidden_dim)

        # for Max-Pooling sublayer
        self.project_graph_pos = nn.Linear(self.embed_dim, self.embed_dim, bias=False)
        self.project_graph_node = nn.Linear(self.embed_dim, self.embed_dim, bias=False)
        self.project_node_pos = nn.Linear(self.embed_dim, self.embed_dim, bias=False)
        self.project_node_node = nn.Linear(self.embed_dim, self.embed_dim, bias=False)

        # for feed-forward aggregation (FFA)sublayer
        self.value_head = MLP(
            input_dim=2 * self.n_heads,
            output_dim=1,
            num_neurons=[32, 32],
            dropout_probs=[0.05, 0.00],
        )

    def forward(self, td: TensorDict, final_h: Tensor, final_p: Tensor) -> Tensor:
        """Compute the logits of the removing a node pair from the current solution

        Args:
            td: TensorDict with the current environment state
            final_h: final NFE embeddings
            final_p: final pfe embeddings
        """

        batch_size, graph_size, dim = final_h.size()

        # Max-Pooling sublayer
        h_node_refined = self.project_node_node(final_h) + self.project_graph_node(
            final_h.max(1)[0]
        )[:, None, :].expand(batch_size, graph_size, dim)
        h_pos_refined = self.project_node_pos(final_p) + self.project_graph_pos(
            final_p.max(1)[0]
        )[:, None, :].expand(batch_size, graph_size, dim)

        # MHC sublayer
        compatibility = torch.zeros(
            (batch_size, graph_size, graph_size, self.n_heads * 2),
            device=h_node_refined.device,
        )
        compatibility[:, :, :, : self.n_heads] = self.compater_pos(h_pos_refined).permute(
            1, 2, 3, 0
        )
        compatibility[:, :, :, self.n_heads :] = self.compater_node(
            h_node_refined
        ).permute(1, 2, 3, 0)

        # FFA sublater
        return self.value_head(self.norm_factor * compatibility).squeeze(-1)


class CriticDecoder(nn.Module):
    def __init__(self, input_dim: int) -> None:
        super().__init__()
        self.input_dim = input_dim

        self.project_graph = nn.Linear(self.input_dim, self.input_dim, bias=False)
        self.project_node = nn.Linear(self.input_dim, self.input_dim, bias=False)

        self.MLP = MLP(
            input_dim=input_dim,
            output_dim=1,
            num_neurons=[input_dim, input_dim // 2],
            dropout_probs=[0.05, 0.0],
        )

    def forward(self, x: torch.Tensor, hidden=None) -> torch.Tensor:
        # h_wave: (batch_size, graph_size+1, input_size)
        mean_pooling = x.mean(1)  # mean Pooling (batch_size, input_size)
        graph_feature: torch.Tensor = self.project_graph(mean_pooling)[
            :, None, :
        ]  # (batch_size, 1, input_dim/2)
        node_feature: torch.Tensor = self.project_node(
            x
        )  # (batch_size, graph_size+1, input_dim/2)

        # pass through value_head, get estimated value
        fusion = node_feature + graph_feature.expand_as(
            node_feature
        )  # (batch_size, graph_size+1, input_dim/2)

        value = self.MLP(fusion.mean(1))

        return value



LKH_SCALING_FACTOR = 100_000
ORTOOLS_SCALING_FACTOR = 100_000
PYVRP_SCALING_FACTOR = 1_000


ROUTEFINDER2LKH = {
    "CVRP": "CVRP",
    "OVRP": "OVRP",
    "OVRPB": None,  # Issue: don't know
    "OVRPBL": None,  # Issue: distance limits
    "OVRPBLTW": None,  # Issue: distance limits
    "OVRPBTW": None,  # Issue: service times don't work in VRPBTW
    "OVRPL": "OVRP",
    "OVRPLTW": "CVRPTW",
    "OVRPMB": "VRPMPD",
    "OVRPMBL": "VRPMPD",
    "OVRPMBTW": "VRPMPDTW",
    "OVRPMBLTW": "VRPMPDTW",  # Issue: distance limits
    "OVRPTW": "CVRPTW",
    "VRPB": None,  # Issue: don't know: linehaul after backhaul
    "VRPBL": None,
    "VRPBLTW": None,  # Issue: service times don't work in VRPBTW
    "VRPBTW": None,  # Issue: service times don't work in VRPBTW
    "VRPL": "DCVRP",
    "VRPLTW": "CVRPTW",  # I don't think that limits get respected
    "VRPMB": "VRPMPD",
    "VRPMBL": "VRPMPD",  # I don't think that limits get respected
    "VRPMBTW": "VRPMPDTW",
    "VRPMBLTW": None,  # Issue: don't know
    "VRPTW": "CVRPTW",
}

LKH_VARIANTS = [
    "CVRP",
    "OVRP",
    "CVRPTW",
    "DCVRP",
    "VRPB",
    "VRPBTW",
    "VRPMPD",
    "VRPMPDTW",
]



import torch.nn as nn

from rl4co.envs import RL4COEnvBase
from rl4co.models.nn.graph.attnnet import MultiHeadAttentionLayer
from rl4co.models.rl import n_step_PPO
from rl4co.models.rl.common.critic import CriticNetwork
from rl4co.models.zoo.n2s.decoder import CriticDecoder
from rl4co.models.zoo.n2s.policy import N2SPolicy
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class N2S(n_step_PPO):
    """N2S Model based on n_step Proximal Policy Optimization (PPO) with an N2S model policy.
    We default to the N2S model policy and the improvement Critic Network.

    Args:
        env: Environment to use for the algorithm
        policy: Policy to use for the algorithm
        critic: Critic to use for the algorithm
        policy_kwargs: Keyword arguments for policy
        critic_kwargs: Keyword arguments for critic
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: nn.Module = None,
        critic: CriticNetwork = None,
        policy_kwargs: dict = {},
        critic_kwargs: dict = {},
        **kwargs,
    ):
        if policy is None:
            policy = N2SPolicy(env_name=env.name, **policy_kwargs)

        if critic is None:
            embed_dim = (
                policy_kwargs["embed_dim"] if "embed_dim" in policy_kwargs else 128
            )  # the critic's embed_dim must be as policy's

            encoder = MultiHeadAttentionLayer(
                embed_dim,
                critic_kwargs["num_heads"] if "num_heads" in critic_kwargs else 4,
                critic_kwargs["feedforward_hidden"]
                if "feedforward_hidden" in critic_kwargs
                else 128,
                critic_kwargs["normalization"]
                if "normalization" in critic_kwargs
                else "layer",
                bias=False,
            )
            value_head = CriticDecoder(embed_dim)

            critic = CriticNetwork(
                encoder=encoder,
                value_head=value_head,
                customized=True,
            )

        super().__init__(env, policy, critic, **kwargs)



import math

from typing import Callable

import torch

from tensordict.tensordict import TensorDict
from torch.distributions import Uniform

from rl4co.envs.common.utils import Generator, get_sampler
from rl4co.utils.ops import get_distance_matrix
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class FLPGenerator(Generator):
    """Data generator for the Facility Location Problem (FLP).

    Args:
        num_loc: number of locations in the FLP
        min_loc: minimum value for the location coordinates
        max_loc: maximum value for the location coordinates
        loc_distribution: distribution for the location coordinates

    Returns:
        A TensorDict with the following keys:
            locs [batch_size, num_loc, 2]: locations
            orig_distances [batch_size, num_loc, num_loc]: original distances between locations
            distances [batch_size, num_loc]: the current minimum distance rom each location to the chosen locations
            chosen [batch_size, num_loc]: indicators of chosen locations
            to_choose [batch_size, 1]: number of locations to choose in the FLP
    """

    def __init__(
        self,
        num_loc: int = 100,
        min_loc: float = 0.0,
        max_loc: float = 1.0,
        loc_distribution: int | float | str | type | Callable = Uniform,
        to_choose: int = 10,
        **kwargs,
    ):
        self.num_loc = num_loc
        self.min_loc = min_loc
        self.max_loc = max_loc
        self.to_choose = to_choose

        # Location distribution
        if kwargs.get("loc_sampler", None) is not None:
            self.loc_sampler = kwargs["loc_sampler"]
        else:
            self.loc_sampler = get_sampler(
                "loc", loc_distribution, min_loc, max_loc, **kwargs
            )

    def _generate(self, batch_size) -> TensorDict:
        # Sample locations
        locs = self.loc_sampler.sample((*batch_size, self.num_loc, 2))
        distances = get_distance_matrix(locs)
        max_dist = math.sqrt(2) * (self.max_loc - self.min_loc)

        return TensorDict(
            {
                "locs": locs,
                "orig_distances": distances,
                "distances": torch.full(
                    (*batch_size, self.num_loc), max_dist, dtype=torch.float
                ),
                "chosen": torch.zeros(*batch_size, self.num_loc, dtype=torch.bool),
                "to_choose": torch.ones(*batch_size, dtype=torch.long) * self.to_choose,
            },
            batch_size=batch_size,
        )



from functools import lru_cache
from typing import Optional

import torch

from einops import rearrange
from tensordict import TensorDict
from torch import Tensor


def _batchify_single(x: Tensor | TensorDict, repeats: int) -> Tensor | TensorDict:
    """Same as repeat on dim=0 for Tensordicts as well"""
    s = x.shape
    return x.expand(repeats, *s).contiguous().view(s[0] * repeats, *s[1:])


def batchify(x: Tensor | TensorDict, shape: tuple | int) -> Tensor | TensorDict:
    """Same as `einops.repeat(x, 'b ... -> (b r) ...', r=repeats)` but ~1.5x faster and supports TensorDicts.
    Repeats batchify operation `n` times as specified by each shape element.
    If shape is a tuple, iterates over each element and repeats that many times to match the tuple shape.

    Example:
    >>> x.shape: [a, b, c, ...]
    >>> shape: [a, b, c]
    >>> out.shape: [a*b*c, ...]
    """
    shape = [shape] if isinstance(shape, int) else shape
    for s in reversed(shape):
        x = _batchify_single(x, s) if s > 0 else x
    return x


def _unbatchify_single(x: Tensor | TensorDict, repeats: int) -> Tensor | TensorDict:
    """Undoes batchify operation for Tensordicts as well"""
    s = x.shape
    return x.view(repeats, s[0] // repeats, *s[1:]).permute(1, 0, *range(2, len(s) + 1))


def unbatchify(x: Tensor | TensorDict, shape: tuple | int) -> Tensor | TensorDict:
    """Same as `einops.rearrange(x, '(r b) ... -> b r ...', r=repeats)` but ~2x faster and supports TensorDicts
    Repeats unbatchify operation `n` times as specified by each shape element
    If shape is a tuple, iterates over each element and unbatchifies that many times to match the tuple shape.

    Example:
    >>> x.shape: [a*b*c, ...]
    >>> shape: [a, b, c]
    >>> out.shape: [a, b, c, ...]
    """
    shape = [shape] if isinstance(shape, int) else shape
    for s in reversed(
        shape
    ):  # we need to reverse the shape to unbatchify in the right order
        x = _unbatchify_single(x, s) if s > 0 else x
    return x


def gather_by_index(src, idx, dim=1, squeeze=True):
    """Gather elements from src by index idx along specified dim

    Example:
    >>> src: shape [64, 20, 2]
    >>> idx: shape [64, 3)] # 3 is the number of idxs on dim 1
    >>> Returns: [64, 3, 2]  # get the 3 elements from src at idx
    """
    expanded_shape = list(src.shape)
    expanded_shape[dim] = -1
    idx = idx.view(idx.shape + (1,) * (src.dim() - idx.dim())).expand(expanded_shape)
    squeeze = idx.size(dim) == 1 and squeeze
    return src.gather(dim, idx).squeeze(dim) if squeeze else src.gather(dim, idx)


def unbatchify_and_gather(x: Tensor, idx: Tensor, n: int):
    """first unbatchify a tensor by n and then gather (usually along the unbatchified dimension)
    by the specified index
    """
    x = unbatchify(x, n)
    return gather_by_index(x, idx, dim=idx.dim())


def get_distance(x: Tensor, y: Tensor):
    """Euclidean distance between two tensors of shape `[..., n, dim]`"""
    return (x - y).norm(p=2, dim=-1)


def get_tour_length(ordered_locs):
    """Compute the total tour distance for a batch of ordered tours.
    Computes the L2 norm between each pair of consecutive nodes in the tour and sums them up.

    Args:
        ordered_locs: Tensor of shape [batch_size, num_nodes, 2] containing the ordered locations of the tour
    """
    ordered_locs_next = torch.roll(ordered_locs, -1, dims=-2)
    return get_distance(ordered_locs_next, ordered_locs).sum(-1)


def get_distance_matrix(locs: Tensor):
    """Compute the euclidean distance matrix for the given coordinates.

    Args:
        locs: Tensor of shape [..., n, dim]
    """
    distance = (locs[..., :, None, :] - locs[..., None, :, :]).norm(p=2, dim=-1)
    return distance


def calculate_entropy(logprobs: Tensor):
    """Calculate the entropy of the log probabilities distribution
    logprobs: Tensor of shape [batch, decoder_steps, num_actions]
    """
    logprobs = torch.nan_to_num(logprobs, nan=0.0)
    entropy = -(logprobs.exp() * logprobs).sum(dim=-1)  # [batch, decoder steps]
    entropy = entropy.sum(dim=1)  # [batch] -- sum over decoding steps
    assert entropy.isfinite().all(), "Entropy is not finite"
    return entropy


# TODO: modularize inside the envs
def get_num_starts(td, env_name=None):
    """Returns the number of possible start nodes for the environment based on the action mask"""
    num_starts = td["action_mask"].shape[-1]
    if env_name == "pdp":
        num_starts = (
            num_starts - 1
        ) // 2  # only half of the nodes (i.e. pickup nodes) can be start nodes
    elif env_name in ["cvrp", "cvrptw", "sdvrp", "mtsp", "op", "pctsp", "spctsp"]:
        num_starts = num_starts - 1  # depot cannot be a start node

    return num_starts


def select_start_nodes(td, env, num_starts):
    """Node selection strategy as proposed in POMO (Kwon et al. 2020)
    and extended in SymNCO (Kim et al. 2022).
    Selects different start nodes for each batch element

    Args:
        td: TensorDict containing the data. We may need to access the available actions to select the start nodes
        env: Environment may determine the node selection strategy
        num_starts: Number of nodes to select. This may be passed when calling the policy directly. See :class:`rl4co.models.AutoregressiveDecoder`
    """
    num_loc = env.generator.num_loc if hasattr(env.generator, "num_loc") else 0xFFFFFFFF
    if env.name in ["tsp", "atsp", "flp", "mcp"]:
        selected = (
            torch.arange(num_starts, device=td.device).repeat_interleave(td.shape[0])
            % num_loc
        )
    elif env.name in ["jssp", "fjsp"]:
        raise NotImplementedError("Multistart not yet supported for FJSP/JSSP")
    else:
        # Environments with depot: we do not select the depot as a start node
        selected = (
            torch.arange(num_starts, device=td.device).repeat_interleave(td.shape[0])
            % num_loc
            + 1
        )
        if env.name == "op":
            if (td["action_mask"][..., 1:].float().sum(-1) < num_starts).any():
                # for the orienteering problem, we may have some nodes that are not available
                # so we need to resample from the distribution of available nodes
                selected = (
                    torch.multinomial(
                        td["action_mask"][..., 1:].float(), num_starts, replacement=True
                    )
                    + 1
                )  # re-add depot index
                selected = rearrange(selected, "b n -> (n b)")
    return selected


def get_best_actions(actions, max_idxs):
    actions = unbatchify(actions, max_idxs.shape[0])
    return actions.gather(0, max_idxs[..., None, None])


def sparsify_graph(cost_matrix: Tensor, k_sparse: Optional[int] = None, self_loop=False):
    """Generate a sparsified graph for the cost_matrix by selecting k edges with the lowest cost for each node.

    Args:
        cost_matrix: Tensor of shape [m, n]
        k_sparse: Number of edges to keep for each node. Defaults to max(n//5, 10) if not provided.
        self_loop: Include self-loop edges in the generated graph when m==n. Defaults to False.
    """
    m, n = cost_matrix.shape
    k_sparse = max(n // 5, 10) if k_sparse is None else k_sparse

    # fill diagonal value with +inf to exclude them from topk results
    if not self_loop and m == n:
        # k_sparse should not exceed n-1 in this occasion
        k_sparse = min(k_sparse, n - 1)
        cost_matrix.fill_diagonal_(torch.inf)

    # select top-k edges with least cost
    topk_values, topk_indices = torch.topk(
        cost_matrix, k=k_sparse, dim=-1, largest=False, sorted=False
    )

    # generate PyG-compatiable edge_index
    edge_index_u = torch.repeat_interleave(
        torch.arange(m, device=cost_matrix.device), topk_indices.shape[1]
    )
    edge_index_v = topk_indices.flatten()
    edge_index = torch.stack([edge_index_u, edge_index_v])

    edge_attr = topk_values.flatten().unsqueeze(-1)
    return edge_index, edge_attr


@lru_cache(5)
def get_full_graph_edge_index(num_node: int, self_loop=False) -> Tensor:
    adj_matrix = torch.ones(num_node, num_node)
    if not self_loop:
        adj_matrix.fill_diagonal_(0)
    edge_index = torch.permute(torch.nonzero(adj_matrix), (1, 0))
    return edge_index


def adj_to_pyg_edge_index(adj: Tensor) -> Tensor:
    """transforms an adjacency matrix (boolean) to a Tensor with the respective edge
    indices (in the format required by the pytorch geometric module).

    :param Tensor adj: shape=(bs, num_nodes, num_nodes)
    :return Tensor: shape=(2, num_edges)
    """
    assert adj.size(1) == adj.size(2), "only symmetric adjacency matrices are supported"
    num_nodes = adj.size(1)
    # (num_edges, 3)
    edge_idx = adj.nonzero()
    batch_idx = edge_idx[:, 0] * num_nodes
    # PyG expects a "single, flat graph", in which the graphs of the batch are not connected.
    # Therefore, add the batch_idx to edge_idx to have unique indices
    flat_edge_idx = edge_idx[:, 1:] + batch_idx[:, None]
    # (2, num_edges)
    flat_edge_idx = torch.permute(flat_edge_idx, (1, 0))
    return flat_edge_idx


def sample_n_random_actions(td: TensorDict, n: int):
    """Helper function to sample n random actions from available actions. If
    number of valid actions is less then n, we sample with replacement from the
    valid actions
    """
    action_mask = td["action_mask"]
    # check whether to use replacement or not
    n_valid_actions = torch.sum(action_mask[:, 1:], 1).min()
    if n_valid_actions < n:
        replace = True
    else:
        replace = False
    ps = torch.rand((action_mask.shape))
    ps[~action_mask] = -torch.inf
    ps = torch.softmax(ps, dim=1)
    selected = torch.multinomial(ps, n, replacement=replace).squeeze(1)
    selected = rearrange(selected, "b n -> (n b)")
    return selected.to(td.device)


def cartesian_to_polar(cartesian: torch.Tensor, origin: Optional[torch.Tensor] = None):
    """Convert Cartesian coordinates to polar coordinates.

    Args:
        cartesian: Tensor of shape [..., 2] containing Cartesian coordinates (x, y)
        origin: Optional origin point to subtract from coordinates before conversion
    """

    if origin is not None:
        cartesian = cartesian - origin
    x, y = cartesian[..., 0], cartesian[..., 1]
    rho = torch.norm(cartesian, dim=-1)
    theta = torch.atan2(y, x)
    polar = torch.stack((rho, theta), dim=-1)
    return polar


def select_start_nodes_by_distance(td, env, num_starts, exclude_depot=True):
    """Select start nodes based on their distance from the origin."""
    polar_locs = td.get("polar_locs", None)
    if polar_locs is None:
        radius = torch.norm(td["locs"], dim=-1)
    else:
        radius = polar_locs[..., 0]
    _, node_index = torch.topk(
        radius, k=num_starts + 1, dim=-1, sorted=True, largest=False
    )
    selected_nodes = node_index[:, 1:] if exclude_depot else node_index[:, :-1]
    return rearrange(selected_nodes, "b n -> (n b)")


def batched_scatter_sum(src, idx):
    """Performs a batched scatter and sum operation on the source tensor using the provided indices.

    Parameters:
        src (Tensor): A tensor of shape [batch_size, N, h].
                      Contains the data to be scattered and summed.
        idx (Tensor): A tensor of shape [batch_size, M, K] with zero-padding.
                      Each non-zero element in idx represents an index (offset by 1)
                      into src. A zero value indicates a padded (invalid) index.
    
    Returns:
        Tensor: A tensor of shape [batch_size, M, h] where for each batch and each index j,
                the output is computed as:
                    Output[batch, j] = sum(src[batch, k - 1] for k in idx[batch, j] if k != 0)
                The subtraction of 1 is applied because 0 is used as the padding value.

    Details:
        - A temporary target tensor (tgt) of shape [batch_size, N+1, h] is created,
          where tgt[:, 1:] is populated with src.
        - The function reshapes idx to gather the corresponding values and then reshapes
          the result back to [batch_size, M, K, h] before summing over the scattering dimension.    
    """
    bs, N, h = src.shape
    bs, M, K = idx.shape
    tgt = torch.zeros(bs, N + 1, h, device=src.device)
    tgt[:, 1:] = src
    tgt = gather_by_index(tgt, idx.long().reshape(bs, -1), squeeze=False)
    return tgt.reshape(bs, M, K, h).sum(-2)    


from typing import IO, Any, Optional, cast

import torch
import torch.nn as nn

from lightning.fabric.utilities.types import _MAP_LOCATION_TYPE, _PATH
from lightning.pytorch.core.saving import _load_from_checkpoint
from tensordict import TensorDict
from typing_extensions import Self

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl.common.base import RL4COLitModule
from rl4co.models.rl.common.utils import RewardScaler
from rl4co.models.rl.reinforce.baselines import REINFORCEBaseline, get_reinforce_baseline
from rl4co.utils.lightning import get_lightning_device
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class REINFORCE(RL4COLitModule):
    """REINFORCE algorithm, also known as policy gradients.
    See superclass `RL4COLitModule` for more details.

    Args:
        env: Environment to use for the algorithm
        policy: Policy to use for the algorithm
        baseline: REINFORCE baseline
        baseline_kwargs: Keyword arguments for baseline. Ignored if baseline is not a string
        **kwargs: Keyword arguments passed to the superclass
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: nn.Module,
        baseline: REINFORCEBaseline | str = "rollout",
        baseline_kwargs: dict = {},
        reward_scale: str = None,
        **kwargs,
    ):
        super().__init__(env, policy, **kwargs)

        self.save_hyperparameters(logger=False)

        if baseline == "critic":
            log.warning(
                "Using critic as baseline. If you want more granular support, use the A2C module instead."
            )

        if isinstance(baseline, str):
            baseline = get_reinforce_baseline(baseline, **baseline_kwargs)
        else:
            if baseline_kwargs != {}:
                log.warning("baseline_kwargs is ignored when baseline is not a string")
        self.baseline = baseline
        self.advantage_scaler = RewardScaler(reward_scale)

    def shared_step(
        self, batch: Any, batch_idx: int, phase: str, dataloader_idx: int = None
    ):
        td = self.env.reset(batch)
        # Perform forward pass (i.e., constructing solution and computing log-likelihoods)
        out = self.policy(td, self.env, phase=phase, select_best=phase != "train")

        # Compute loss
        if phase == "train":
            out = self.calculate_loss(td, batch, out)

        metrics = self.log_metrics(out, phase, dataloader_idx=dataloader_idx)
        return {"loss": out.get("loss", None), **metrics}

    def calculate_loss(
        self,
        td: TensorDict,
        batch: TensorDict,
        policy_out: dict,
        reward: Optional[torch.Tensor] = None,
        log_likelihood: Optional[torch.Tensor] = None,
    ):
        """Calculate loss for REINFORCE algorithm.

        Args:
            td: TensorDict containing the current state of the environment
            batch: Batch of data. This is used to get the extra loss terms, e.g., REINFORCE baseline
            policy_out: Output of the policy network
            reward: Reward tensor. If None, it is taken from `policy_out`
            log_likelihood: Log-likelihood tensor. If None, it is taken from `policy_out`
        """
        # Extra: this is used for additional loss terms, e.g., REINFORCE baseline
        extra = batch.get("extra", None)
        reward = reward if reward is not None else policy_out["reward"]
        log_likelihood = (
            log_likelihood if log_likelihood is not None else policy_out["log_likelihood"]
        )

        # REINFORCE baseline
        bl_val, bl_loss = (
            self.baseline.eval(td, reward, self.env) if extra is None else (extra, 0)
        )

        # Main loss function
        advantage = reward - bl_val  # advantage = reward - baseline
        advantage = self.advantage_scaler(advantage)
        reinforce_loss = -(advantage * log_likelihood).mean()
        loss = reinforce_loss + bl_loss
        policy_out.update(
            {
                "loss": loss,
                "reinforce_loss": reinforce_loss,
                "bl_loss": bl_loss,
                "bl_val": bl_val,
            }
        )
        return policy_out

    def post_setup_hook(self, stage="fit"):
        # Make baseline taking model itself and train_dataloader from model as input
        self.baseline.setup(
            self.policy,
            self.env,
            batch_size=self.val_batch_size,
            device=get_lightning_device(self),
            dataset_size=self.data_cfg["val_data_size"],
        )

    def on_train_epoch_end(self):
        """Callback for end of training epoch: we evaluate the baseline"""
        self.baseline.epoch_callback(
            self.policy,
            env=self.env,
            batch_size=self.val_batch_size,
            device=get_lightning_device(self),
            epoch=self.current_epoch,
            dataset_size=self.data_cfg["val_data_size"],
        )
        # Need to call super() for the dataset to be reset
        super().on_train_epoch_end()

    def wrap_dataset(self, dataset):
        """Wrap dataset from baseline evaluation. Used in greedy rollout baseline"""
        return self.baseline.wrap_dataset(
            dataset,
            self.env,
            batch_size=self.val_batch_size,
            device=get_lightning_device(self),
        )

    def set_decode_type_multistart(self, phase: str):
        """Set decode type to `multistart` for train, val and test in policy.
        For example, if the decode type is `greedy`, it will be set to `multistart_greedy`.

        Args:
            phase: Phase to set decode type for. Must be one of `train`, `val` or `test`.
        """
        attribute = f"{phase}_decode_type"
        attr_get = getattr(self.policy, attribute)
        # If does not exist, log error
        if attr_get is None:
            log.error(f"Decode type for {phase} is None. Cannot prepend `multistart_`.")
            return
        elif "multistart" in attr_get:
            return
        else:
            setattr(self.policy, attribute, f"multistart_{attr_get}")

    @classmethod
    def load_from_checkpoint(
        cls,
        checkpoint_path: _PATH | IO,
        map_location: _MAP_LOCATION_TYPE = None,
        hparams_file: Optional[_PATH] = None,
        strict: bool = False,
        load_baseline: bool = True,
        **kwargs: Any,
    ) -> Self:
        """Load model from checkpoint/

        Note:
            This is a modified version of `load_from_checkpoint` from `pytorch_lightning.core.saving`.
            It deals with matching keys for the baseline by first running setup
        """

        if strict:
            log.warning("Setting strict=False for loading model from checkpoint.")
            strict = False

        # Do not use strict
        loaded = _load_from_checkpoint(
            cls,
            checkpoint_path,
            map_location,
            hparams_file,
            strict,
            **kwargs,
        )

        # Load baseline state dict
        if load_baseline:
            # setup baseline first
            loaded.setup()
            loaded.post_setup_hook()
            # load baseline state dict
            state_dict = torch.load(
                checkpoint_path, map_location=map_location, weights_only=False
            )["state_dict"]
            # get only baseline parameters
            state_dict = {k: v for k, v in state_dict.items() if "baseline" in k}
            state_dict = {k.replace("baseline.", "", 1): v for k, v in state_dict.items()}
            loaded.baseline.load_state_dict(state_dict)

        return cast(Self, loaded)



import torch
import numpy as np
import matplotlib.pyplot as plt

from matplotlib import cm, colormaps
from tensordict.tensordict import TensorDict

from rl4co.utils.ops import gather_by_index
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def render(td: TensorDict, actions=None, ax=None):
    raise NotImplementedError



import logging

from typing import Tuple

import torch

from tensordict import TensorDict
from torch import Size, Tensor

from rl4co.envs.scheduling.fjsp import INIT_FINISH

logger = logging.getLogger(__name__)


def get_op_features(td: TensorDict):
    return torch.stack((td["lbs"], td["is_ready"], td["num_eligible"]), dim=-1)


def cat_and_norm_features(
    td: TensorDict, feats: list[str], time_feats: list[str], norm_const: int
):
    # logger.info(f"will scale the features {','.join(time_feats)} with a constant ({norm_const})")
    feature_list = []
    for feat in feats:
        if feat in time_feats:
            feature_list.append(td[feat] / norm_const)
        else:
            feature_list.append(td[feat])

    return torch.stack(feature_list, dim=-1).to(torch.float32)


def view(
    tensor: Tensor,
    idx: Tuple[Tensor],
    pad_mask: Tensor,
    new_shape: Size | list[int],
    pad_value: float | int,
):
    # convert mask specifying which entries are padded into mask specifying which entries to keep
    mask = ~pad_mask
    new_view = tensor.new_full(size=new_shape, fill_value=pad_value)
    new_view[idx] = tensor[mask]
    return new_view


def _get_idx_for_job_op_view(td: TensorDict) -> tuple:
    bs, _, n_total_ops = td["job_ops_adj"].shape
    # (bs, ops)
    batch_idx = torch.arange(bs, device=td.device).repeat_interleave(n_total_ops)
    batch_idx = batch_idx.reshape(bs, -1)
    # (bs, ops)
    ops_job_map = td["ops_job_map"]
    # (bs, ops)
    ops_sequence_order = td["ops_sequence_order"]
    # (bs*n_ops_max, 3)
    idx = (
        torch.stack((batch_idx, ops_job_map, ops_sequence_order), dim=-1)
        .to(torch.long)
        .flatten(0, 1)
    )
    # (bs, n_ops_max)
    mask = ~td["pad_mask"]
    # (total_ops_in_batch, 3)
    idx = idx[mask.flatten(0, 1)]
    b, j, o = map(lambda x: x.squeeze(1), idx.chunk(3, dim=-1))
    return b, j, o


def get_job_op_view(td: TensorDict, keys: list[str] = [], pad_value: float | int = 0):
    """This function reshapes all tensors of the tensordict from a flat operations-only view
    to a nested job-operation view and creates a new tensordict from it.
    :param _type_ td: tensordict
    :return _type_: dict
    """
    # ============= Prepare the new index =============
    bs, num_jobs, _ = td["job_ops_adj"].shape
    max_ops_per_job = int(td["job_ops_adj"].sum(-1).max())
    idx = _get_idx_for_job_op_view(td)
    new_shape = Size((bs, num_jobs, max_ops_per_job))
    pad_mask = td["pad_mask"]
    # ==============================================

    # due to special structure, processing times are treated seperately
    if "proc_times" in keys:
        keys.remove("proc_times")
    # reshape processing times; (bs, ma, ops) -> (bs, ma, jobs, ops_per_job)
    new_proc_times_view = view(
        td["proc_times"].permute(0, 2, 1), idx, pad_mask, new_shape, pad_value
    ).permute(0, 3, 1, 2)

    # add padding mask if not in keys
    if "pad_mask" not in keys:
        keys.append("pad_mask")

    new_views = dict(
        map(lambda key: (key, view(td[key], idx, pad_mask, new_shape)), keys)
    )

    # update tensordict clone with reshaped tensors
    return {"proc_times": new_proc_times_view, **new_views}


def blockify(td, tensor: Tensor, pad_value: float | int = 0):
    assert len(tensor.shape) in [
        2,
        3,
    ], "blockify only supports tensors of shape (bs, seq, (d)), where the feature dim d is optional"
    # get the size of the blockified tensor
    bs, _, *d = tensor.shape
    num_jobs = td["job_ops_adj"].size(1)
    max_ops_per_job = int(td["job_ops_adj"].sum(-1).max())
    new_shape = Size((bs, num_jobs, max_ops_per_job, *d))
    # get indices of valid entries of blockified tensor
    idx = _get_idx_for_job_op_view(td)
    pad_mask = td["pad_mask"]
    # create the blockified view
    new_view_tensor = view(tensor, idx, pad_mask, new_shape, pad_value)
    return new_view_tensor


def unblockify(
    td: TensorDict, tensor: Tensor, mask: Tensor = None, pad_value: float | int = 0
):
    assert len(tensor.shape) in [
        3,
        4,
    ], "blockify only supports tensors of shape (bs, nb, s, (d)), where the feature dim d is optional"
    # get the size of the blockified tensor
    bs, _, _, *d = tensor.shape
    n_ops_per_batch = td["job_ops_adj"].sum((1, 2)).unsqueeze(1)  # (bs)
    seq_len = int(n_ops_per_batch.max())
    new_shape = Size((bs, seq_len, *d))

    # create the mask to gather then entries of the blockified tensor. NOTE that only by
    # blockifying the original pad_mask
    pad_mask = td["pad_mask"]
    pad_mask = blockify(td, pad_mask, True)

    # get indices of valid entrie in flat matrix
    b = torch.arange(bs, device=td.device).repeat_interleave(seq_len).reshape(bs, seq_len)
    i = torch.arange(seq_len, device=td.device)[None].repeat(bs, 1)
    idx = tuple(map(lambda x: x[i < n_ops_per_batch], (b, i)))
    # create view
    new_tensor = view(tensor, idx, pad_mask, new_shape, pad_value=pad_value)
    return new_tensor


def first_diff(x: Tensor, dim: int):
    shape = x.shape
    shape = (*shape[:dim], 1, *shape[dim + 1 :])
    seq_cutoff = x.index_select(dim, torch.arange(x.size(dim) - 1, device=x.device))
    first_diff_seq = x - torch.cat((seq_cutoff.new_zeros(*shape), seq_cutoff), dim=dim)
    return first_diff_seq


def spatial_encoding(td: TensorDict):
    """We use a spatial encoing as proposed in GraphFormer (https://arxiv.org/abs/2106.05234)
    The spatial encoding in GraphFormer determines the distance of the shortest path between and
    nodes i and j and uses a special value for node pairs that cannot be connected at all.
    For any two operations i<j of the same job, we determine the number of operations to be completet
    when starting at i before j can be started (e.g. i=3 and j=5 -> e=2) and for i>j the negative number of
    operations that starting from j, have been completet before arriving at i (e.g. i=5 j=3 -> e=-2).
    For i=j we set e=0 as well as for operations of different jobs.

    :param torch.Tensor[bs, n_ops] ops_job_map: tensor specifying the index of its corresponding job
    :return torch.Tensor[bs, n_ops, n_ops]: length of shortest path between any two operations
    """
    bs, _, n_total_ops = td["job_ops_adj"].shape
    max_ops_per_job = int(td["job_ops_adj"].sum(-1).max())
    ops_job_map = td["ops_job_map"]
    pad_mask = td["pad_mask"]

    same_job = (ops_job_map[:, None] == ops_job_map[..., None]).to(torch.int32)
    # mask padded
    same_job[pad_mask.unsqueeze(2).expand_as(same_job)] = 0
    same_job[pad_mask.unsqueeze(1).expand_as(same_job)] = 0
    # take upper triangular of same_job and set diagonal to zero for counting purposes
    upper_tri = torch.triu(same_job) - torch.diag(
        torch.ones(n_total_ops, device=td.device)
    )[None].expand_as(same_job)
    # cumsum and masking of operations that do not belong to the same job
    num_jumps = upper_tri.cumsum(2) * upper_tri
    # mirror the matrix
    num_jumps = num_jumps + num_jumps.transpose(1, 2)
    # NOTE: shifted this logic into the spatial encoding module
    # num_jumps = num_jumps + (-num_jumps.transpose(1,2))
    assert not torch.any(num_jumps >= max_ops_per_job)
    # special value for ops of different jobs and self-loops
    num_jumps = torch.where(num_jumps == 0, -1, num_jumps)
    self_mask = torch.eye(n_total_ops).repeat(bs, 1, 1).bool()
    num_jumps[self_mask] = 0
    return num_jumps


def calc_lower_bound(td: TensorDict):
    """Here we calculate the lower bound of the operations finish times. In the FJSP case, multiple things need to
    be taken into account due to the usability of the different machines for multiple ops of different jobs:

    1.) Operations may only start once their direct predecessor is finished. We calculate its lower bound by
    adding the minimum possible operation time to this detected start time. However, we cannot use the proc_times
    directly, but need to account for the fact, that machines might still be busy, once an operation can be processed.
    We detect this offset by detecting ops-machine pairs, where the first possible start point of the operation is before
    the machine becomes idle again - Therefore, we add this discrepancy to the proc_time of the respective ops-ma combination

    2.) If an operation has been scheduled, we use its actual finishing time as lower bound. In this case, using the cumulative sum
    of all peedecessors of a job does not make sense, since it is likely to differ from the real finishing time of its direct
    predecessor (its only a lower bound). Therefore, we add the finish time to the cumulative sum of processing time of all
    UNSCHEDULED operations, to obtain the lower bound.
    Making this work is a bit hacky: We compute the first differences of finishing times of those operations scheduled and
    add them to the matrix of processing times, where already processed operations are masked (with zero)


    """

    proc_times = td["proc_times"].clone()  # (bs, ma, ops)
    busy_until = td["busy_until"]  # (bs, ma)
    ops_adj = td["ops_adj"]  # (bs, ops, ops, 2)
    finish_times = td["finish_times"]  # (bs, ops)
    job_ops_adj = td["job_ops_adj"]  # (bs, jobs, ops)
    op_scheduled = td["op_scheduled"].to(torch.float32)  # (bs, ops)

    ############## REGARDING POINT 1 OF DOCSTRING ##############
    # for operations whose immidiate predecessor is scheduled, we can determine its earliest
    # start time by the end time of the predecessor.
    # (bs, num_ops, 1)
    maybe_start_at = torch.bmm(ops_adj[..., 0], finish_times[..., None]).squeeze(2)
    # using the start_time, we can determine if and how long an op needs to wait for a machine to finish
    wait_for_ma_offset = torch.clip(busy_until[..., None] - maybe_start_at[:, None], 0)
    # we add this required waiting time to the respective processing time
    proc_time_plus_wait = torch.where(
        proc_times == 0, proc_times, proc_times + wait_for_ma_offset
    )
    # NOTE get the mean processing time over all eligible machines for lb calulation
    # ops_proc_times = torch.where(proc_times == 0, torch.inf, proc_time_plus_wait).min(1).values)
    ops_proc_times = proc_time_plus_wait.sum(1) / (proc_times.gt(0).sum(1) + 1e-9)
    # mask proc times for already scheduled ops
    ops_proc_times[op_scheduled.to(torch.bool)] = 0

    ############### REGARDING POINT 2 OF DOCSTRING ###################
    # Now we determine all operations that are not scheduled yet (and thus have no finish_time). We will compute the cumulative
    # sum over the processing time to determine the lower bound of unscheduled operations...
    proc_matrix = job_ops_adj
    ops_assigned = proc_matrix * op_scheduled[:, None]
    proc_matrix_not_scheduled = proc_matrix * (
        torch.ones_like(proc_matrix) - op_scheduled[:, None]
    )

    # ...and add the finish_time of the last scheduled operation of the respective job to that. To make this work, using the cumsum logic,
    # we calc the first differences of the finish times and seperate by job.
    # We use the first differences, so that the finish times do not add up during cumulative sum below
    # (bs, num_jobs, num_ops)
    finish_times_1st_diff = ops_assigned * first_diff(
        ops_assigned * finish_times[:, None], 2
    )

    # masking the processing time of scheduled operations and add their finish times instead (first diff thereof)
    lb_end_expand = (
        proc_matrix_not_scheduled * ops_proc_times.unsqueeze(1).expand_as(job_ops_adj)
        + finish_times_1st_diff
    )
    # (bs, max_ops); lower bound finish time per operation using the cumsum logic
    LBs = torch.sum(job_ops_adj * lb_end_expand.cumsum(-1), dim=1)
    # remove nans
    LBs = torch.nan_to_num(LBs, nan=0.0)

    # test
    assert torch.where(
        finish_times != INIT_FINISH, torch.isclose(LBs, finish_times), True
    ).all()

    return LBs


def op_is_ready(td: TensorDict):
    # compare finish times of predecessors with current time step; shape=(b, n_ops_max)
    is_ready = (
        torch.bmm(td["ops_adj"][..., 0], td["finish_times"][..., None]).squeeze(2)
        <= td["time"][:, None]
    )
    # shape=(b, n_ops_max)
    is_scheduled = td["ma_assignment"].sum(1).bool()
    # op is ready for scheduling if it has not been scheduled and its predecessor is finished
    return torch.logical_and(is_ready, ~is_scheduled)


def get_job_ops_mapping(
    start_op_per_job: torch.Tensor, end_op_per_job: torch.Tensor, n_ops_max: int
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Implements a mapping function from operations to jobs

    :param torch.Tensor start_op_per_job: index of first operation of each job
    :param torch.Tensor end_op_per_job: index of last operation of each job
    :return Tuple[torch.Tensor, torch.Tensor]:
        1st.) index mapping (bs, num_ops): [0,0,1,1,1] means that first two operations belong to job 0
        2st.) binary mapping (bs, num_jobs, num_ops): [[1,1,0], [0,0,1]] means that first two operations belong to job 0
    """
    device = end_op_per_job.device
    end_op_per_job = end_op_per_job.clone()

    bs, num_jobs = end_op_per_job.shape

    # in order to avoid shape conflicts, set the end operation id to the id of max_ops (all batches have same #ops)
    end_op_per_job[:, -1] = n_ops_max - 1

    # here we will generate the operations-job mapping:
    # Therefore we first generate a sequence of operation ids and expand it the the size of the mapping matrix:
    # (bs, jobs, max_ops)
    ops_seq_exp = torch.arange(n_ops_max, device=device)[None, None].expand(
        bs, num_jobs, -1
    )
    # (bs, jobs, max_ops)  # expanding start and end operation ids
    end_op_per_job_exp = end_op_per_job[..., None].expand_as(ops_seq_exp)
    start_op_per_job_exp = start_op_per_job[..., None].expand_as(ops_seq_exp)
    # given ids of start and end operations per job, this generates the mapping of ops to jobs
    # (bs, jobs, max_ops)
    ops_job_map = torch.nonzero(
        (ops_seq_exp <= end_op_per_job_exp) & (ops_seq_exp >= start_op_per_job_exp)
    )
    # (bs, max_ops)
    ops_job_map = torch.stack(ops_job_map[:, 1].split(n_ops_max), dim=0)

    # we might also want a binary mapping / adjacency matrix connecting jobs to operations
    # (bs, num_jobs, num_ops)
    ops_job_bin_map = torch.scatter_add(
        input=ops_job_map.new_zeros((bs, num_jobs, n_ops_max)),
        dim=1,
        index=ops_job_map.unsqueeze(1),
        src=ops_job_map.new_ones((bs, num_jobs, n_ops_max)),
    )

    return ops_job_map, ops_job_bin_map



from typing import Callable, Union

from tensordict.tensordict import TensorDict
from torch.distributions import Uniform

from rl4co.envs.common.utils import Generator, get_sampler
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class SHPPGenerator(Generator):
    """Data generator for the Shortest Hamiltonian Path Problem (SHPP).
    Args:
        num_loc: number of locations (customers) in the TSP
        min_loc: minimum value for the location coordinates
        max_loc: maximum value for the location coordinates
        loc_distribution: distribution for the location coordinates

    Returns:
        A TensorDict with the following keys:
            locs [batch_size, num_loc, 2]: locations of each customer
    """

    def __init__(
        self,
        num_loc: int = 20,
        min_loc: float = 0.0,
        max_loc: float = 1.0,
        loc_distribution: Union[int, float, str, type, Callable] = Uniform,
        **kwargs,
    ):
        self.num_loc = num_loc
        self.min_loc = min_loc
        self.max_loc = max_loc

        # Location distribution
        if kwargs.get("loc_sampler", None) is not None:
            self.loc_sampler = kwargs["loc_sampler"]
        else:
            self.loc_sampler = get_sampler(
                "loc", loc_distribution, min_loc, max_loc, **kwargs
            )

    def _generate(self, batch_size) -> TensorDict:
        # Sample locations
        locs = self.loc_sampler.sample((*batch_size, self.num_loc, 2))

        return TensorDict(
            {
                "locs": locs,
            },
            batch_size=batch_size,
        )



from dataclasses import dataclass, fields
from typing import Tuple

import torch
import torch.nn as nn

from einops import rearrange
from tensordict import TensorDict
from torch import Tensor

from rl4co.envs import RL4COEnvBase
from rl4co.models.common.constructive.autoregressive.decoder import AutoregressiveDecoder
from rl4co.models.nn.attention import PointerAttention, PointerAttnMoE
from rl4co.models.nn.env_embeddings import env_context_embedding, env_dynamic_embedding
from rl4co.models.nn.env_embeddings.dynamic import StaticEmbedding
from rl4co.utils.ops import batchify, unbatchify
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


@dataclass
class PrecomputedCache:
    node_embeddings: Tensor
    graph_context: Tensor | float
    glimpse_key: Tensor
    glimpse_val: Tensor
    logit_key: Tensor

    @property
    def fields(self):
        return tuple(getattr(self, x.name) for x in fields(self))

    def batchify(self, num_starts):
        new_embs = []
        for emb in self.fields:
            if isinstance(emb, Tensor) or isinstance(emb, TensorDict):
                new_embs.append(batchify(emb, num_starts))
            else:
                new_embs.append(emb)
        return PrecomputedCache(*new_embs)


class AttentionModelDecoder(AutoregressiveDecoder):
    """
    Auto-regressive decoder based on Kool et al. (2019): https://arxiv.org/abs/1803.08475.
    Given the environment state and the embeddings, compute the logits and sample actions autoregressively until
    all the environments in the batch have reached a terminal state.
    In this case we additionally have a `pre_decoder_hook` method that allows to precompute the embeddings before
    the decoder is called, which saves a lot of computation.


    Args:
        embed_dim: Embedding dimension
        num_heads: Number of attention heads
        env_name: Name of the environment used to initialize embeddings
        context_embedding: Context embedding module
        dynamic_embedding: Dynamic embedding module
        mask_inner: Whether to mask the inner loop
        out_bias_pointer_attn: Whether to use a bias in the pointer attention
        linear_bias: Whether to use a bias in the linear layer
        use_graph_context: Whether to use the graph context
        check_nan: Whether to check for nan values during decoding
        sdpa_fn: scaled_dot_product_attention function
        pointer: Module implementing the pointer logic (defaults to PointerAttention)
        moe_kwargs: Keyword arguments for MoE
    """

    def __init__(
        self,
        embed_dim: int = 128,
        num_heads: int = 8,
        env_name: str = "tsp",
        context_embedding: nn.Module = None,
        dynamic_embedding: nn.Module = None,
        mask_inner: bool = True,
        out_bias_pointer_attn: bool = False,
        linear_bias: bool = False,
        use_graph_context: bool = True,
        check_nan: bool = True,
        sdpa_fn: callable = None,
        pointer: nn.Module = None,
        moe_kwargs: dict = None,
    ):
        super().__init__()

        if isinstance(env_name, RL4COEnvBase):
            env_name = env_name.name
        self.env_name = env_name
        self.embed_dim = embed_dim
        self.num_heads = num_heads

        assert embed_dim % num_heads == 0

        self.context_embedding = (
            env_context_embedding(self.env_name, {"embed_dim": embed_dim})
            if context_embedding is None
            else context_embedding
        )
        self.dynamic_embedding = (
            env_dynamic_embedding(self.env_name, {"embed_dim": embed_dim})
            if dynamic_embedding is None
            else dynamic_embedding
        )
        self.is_dynamic_embedding = (
            False if isinstance(self.dynamic_embedding, StaticEmbedding) else True
        )

        if pointer is None:
            # MHA with Pointer mechanism (https://arxiv.org/abs/1506.03134)
            pointer_attn_class = (
                PointerAttention if moe_kwargs is None else PointerAttnMoE
            )
            pointer = pointer_attn_class(
                embed_dim,
                num_heads,
                mask_inner=mask_inner,
                out_bias=out_bias_pointer_attn,
                check_nan=check_nan,
                sdpa_fn=sdpa_fn,
                moe_kwargs=moe_kwargs,
            )

        self.pointer = pointer

        # For each node we compute (glimpse key, glimpse value, logit key) so 3 * embed_dim
        self.project_node_embeddings = nn.Linear(
            embed_dim, 3 * embed_dim, bias=linear_bias
        )
        self.project_fixed_context = nn.Linear(embed_dim, embed_dim, bias=linear_bias)
        self.use_graph_context = use_graph_context

    def _compute_q(self, cached: PrecomputedCache, td: TensorDict):
        node_embeds_cache = cached.node_embeddings
        graph_context_cache = cached.graph_context

        if td.dim() == 2 and isinstance(graph_context_cache, Tensor):
            graph_context_cache = graph_context_cache.unsqueeze(1)

        step_context = self.context_embedding(node_embeds_cache, td)
        glimpse_q = step_context + graph_context_cache
        # add seq_len dim if not present
        glimpse_q = glimpse_q.unsqueeze(1) if glimpse_q.ndim == 2 else glimpse_q

        return glimpse_q

    def _compute_kvl(self, cached: PrecomputedCache, td: TensorDict):
        glimpse_k_stat, glimpse_v_stat, logit_k_stat = (
            cached.glimpse_key,
            cached.glimpse_val,
            cached.logit_key,
        )
        # Compute dynamic embeddings and add to static embeddings
        glimpse_k_dyn, glimpse_v_dyn, logit_k_dyn = self.dynamic_embedding(td)
        glimpse_k = glimpse_k_stat + glimpse_k_dyn
        glimpse_v = glimpse_v_stat + glimpse_v_dyn
        logit_k = logit_k_stat + logit_k_dyn

        return glimpse_k, glimpse_v, logit_k

    def forward(
        self,
        td: TensorDict,
        cached: PrecomputedCache,
        num_starts: int = 0,
    ) -> Tuple[Tensor, Tensor]:
        """Compute the logits of the next actions given the current state

        Args:
            cache: Precomputed embeddings
            td: TensorDict with the current environment state
            num_starts: Number of starts for the multi-start decoding
        """

        has_dyn_emb_multi_start = self.is_dynamic_embedding and num_starts > 1

        # Handle efficient multi-start decoding
        if has_dyn_emb_multi_start:
            # if num_starts > 0 and we have some dynamic embeddings, we need to reshape them to [B*S, ...]
            # since keys and values are not shared across starts (i.e. the episodes modify these embeddings at each step)
            cached = cached.batchify(num_starts=num_starts)

        elif num_starts > 1:
            td = unbatchify(td, num_starts)

        glimpse_q = self._compute_q(cached, td)
        glimpse_k, glimpse_v, logit_k = self._compute_kvl(cached, td)

        # Compute logits
        mask = td["action_mask"]
        logits = self.pointer(glimpse_q, glimpse_k, glimpse_v, logit_k, mask)

        # Now we need to reshape the logits and mask to [B*S,N,...] is num_starts > 1 without dynamic embeddings
        # note that rearranging order is important here
        if num_starts > 1 and not has_dyn_emb_multi_start:
            logits = rearrange(logits, "b s l -> (s b) l", s=num_starts)
            mask = rearrange(mask, "b s l -> (s b) l", s=num_starts)
        return logits, mask

    def pre_decoder_hook(
        self, td, env, embeddings, num_starts: int = 0
    ) -> Tuple[TensorDict, RL4COEnvBase, PrecomputedCache]:
        """Precompute the embeddings cache before the decoder is called"""
        return td, env, self._precompute_cache(embeddings, num_starts=num_starts)

    def _precompute_cache(
        self, embeddings: torch.Tensor, num_starts: int = 0
    ) -> PrecomputedCache:
        """Compute the cached embeddings for the pointer attention.

        Args:
            embeddings: Precomputed embeddings for the nodes
            num_starts: Number of starts for the multi-start decoding
        """
        # The projection of the node embeddings for the attention is calculated once up front
        (
            glimpse_key_fixed,
            glimpse_val_fixed,
            logit_key_fixed,
        ) = self.project_node_embeddings(embeddings).chunk(3, dim=-1)

        # Optionally disable the graph context from the initial embedding as done in POMO
        if self.use_graph_context:
            graph_context = self.project_fixed_context(embeddings.mean(1))
        else:
            graph_context = 0

        # Organize in a dataclass for easy access
        return PrecomputedCache(
            node_embeddings=embeddings,
            graph_context=graph_context,
            glimpse_key=glimpse_key_fixed,
            glimpse_val=glimpse_val_fixed,
            logit_key=logit_key_fixed,
        )



import math

from typing import Optional, Union

import numpy as np
import scipy
import torch

from tensordict import TensorDict

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl.reinforce.baselines import REINFORCEBaseline
from rl4co.models.zoo.deepaco import DeepACO
from rl4co.models.zoo.gfacs.policy import GFACSPolicy
from rl4co.utils.ops import unbatchify


class GFACS(DeepACO):
    """Implements GFACS: https://arxiv.org/abs/2403.07041.

    Args:
        env: Environment to use for the algorithm
        policy: Policy to use for the algorithm
        baseline: REINFORCE baseline. Defaults to no baseline
        train_with_local_search: Whether to train with local search. Defaults to False.
        policy_kwargs: Keyword arguments for policy
        baseline_kwargs: Keyword arguments for baseline
        alpha_min: Minimum value for alpha. Defaults to 0.5.
        alpha_max: Maximum value for alpha. Defaults to 1.0.
        alpha_flat_epochs: Number of epochs to keep alpha constant. Defaults to 5.
        beta_min: Minimum value for beta. Defaults to 1.0.
        beta_max: Maximum value for beta. Defaults to 1.0.
        beta_flat_epochs: Number of epochs to keep beta constant. Defaults to 5.
        **kwargs: Keyword arguments passed to the superclass
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: Optional[GFACSPolicy] = None,
        baseline: Union[REINFORCEBaseline, str] = "no",
        train_with_local_search: bool = True,
        policy_kwargs: dict = {},
        baseline_kwargs: dict = {},
        alpha_min: float = 0.5,
        alpha_max: float = 1.0,
        alpha_flat_epochs: int = 5,
        beta_min: float = 1.0,
        beta_max: float = 1.0,
        beta_flat_epochs: int = 5,
        **kwargs,
    ):
        if policy is None:
            policy = GFACSPolicy(
                env_name=env.name,
                train_with_local_search=train_with_local_search,
                **policy_kwargs,
            )

        super().__init__(
            env,
            policy,
            baseline,
            train_with_local_search,
            policy_kwargs=policy_kwargs,
            baseline_kwargs=baseline_kwargs,
            **kwargs,  # ls_reward_aug_W is replaced by \alpha in this paper
        )

        self.alpha_min = alpha_min
        self.alpha_max = alpha_max
        self.alpha_flat_epochs = alpha_flat_epochs
        self.beta_min = beta_min
        self.beta_max = beta_max
        self.beta_flat_epochs = beta_flat_epochs

    @property
    def alpha(self) -> float:
        return self.alpha_min + (self.alpha_max - self.alpha_min) * min(
            self.current_epoch
            / (self.trainer.max_epochs - self.alpha_flat_epochs),  # type: ignore
            1.0,
        )

    @property
    def beta(self) -> float:
        return self.beta_min + (self.beta_max - self.beta_min) * min(
            math.log(self.current_epoch + 1)
            / math.log(self.trainer.max_epochs - self.beta_flat_epochs),  # type: ignore
            1.0,
        )

    def calculate_loss(
        self,
        td: TensorDict,
        batch: TensorDict,
        policy_out: dict,
        reward: Optional[torch.Tensor] = None,
        log_likelihood: Optional[torch.Tensor] = None,
    ):
        """Calculate loss for REINFORCE algorithm.

        Args:
            td: TensorDict containing the current state of the environment
            batch: Batch of data. This is used to get the extra loss terms, e.g., REINFORCE baseline
            policy_out: Output of the policy network
            reward: Reward tensor. If None, it is taken from `policy_out`
            log_likelihood: Log-likelihood tensor. If None, it is taken from `policy_out`
        """
        reward = policy_out["reward"]
        n_ants = reward.size(1)  # type: ignore
        advantage = reward - reward.mean(dim=1, keepdim=True)  # type: ignore

        if self.train_with_local_search:
            ls_reward = policy_out["ls_reward"]
            ls_advantage = ls_reward - ls_reward.mean(dim=1, keepdim=True)
            weighted_advantage = advantage * (1 - self.alpha) + ls_advantage * self.alpha
        else:
            weighted_advantage = advantage

        # On-policy loss
        forward_flow = policy_out["log_likelihood"] + policy_out["logZ"].repeat(1, n_ants)
        backward_flow = (
            self.calculate_log_pb_uniform(policy_out["actions"], n_ants)
            + weighted_advantage.detach() * self.beta
        )
        tb_loss = torch.pow(forward_flow - backward_flow, 2).mean()

        # Off-policy loss
        if self.train_with_local_search:
            ls_forward_flow = policy_out["ls_log_likelihood"] + policy_out[
                "ls_logZ"
            ].repeat(1, n_ants)
            ls_backward_flow = (
                self.calculate_log_pb_uniform(policy_out["ls_actions"], n_ants)
                + ls_advantage.detach() * self.beta
            )
            ls_tb_loss = torch.pow(ls_forward_flow - ls_backward_flow, 2).mean()
            tb_loss = tb_loss + ls_tb_loss

        return tb_loss

    def calculate_log_pb_uniform(self, actions: torch.Tensor, n_ants: int):
        match self.env.name:
            case "tsp":
                return math.log(1 / 2 * actions.size(1))
            case "cvrp":
                _a1 = actions.detach().cpu().numpy()
                # shape: (batch, max_tour_length)
                n_nodes = np.count_nonzero(_a1, axis=1)
                _a2 = _a1[:, 1:] - _a1[:, :-1]
                n_routes = np.count_nonzero(_a2, axis=1) - n_nodes
                _a3 = _a1[:, 2:] - _a1[:, :-2]
                n_multinode_routes = np.count_nonzero(_a3, axis=1) - n_nodes
                log_b_p = -scipy.special.gammaln(
                    n_routes + 1
                ) - n_multinode_routes * math.log(2)
                return unbatchify(torch.from_numpy(log_b_p).to(actions.device), n_ants)
            case "op" | "pctsp":
                return math.log(1 / 2)
            case _:
                raise ValueError(f"Unknown environment: {self.env.name}")



import abc

from typing import Any, Tuple

import torch
import torch.nn as nn

from einops import einsum, rearrange
from tensordict import TensorDict
from torch import Tensor

from rl4co.models.common.constructive.autoregressive import AutoregressiveDecoder
from rl4co.models.nn.attention import PointerAttention
from rl4co.models.nn.env_embeddings.context import SchedulingContext
from rl4co.models.nn.env_embeddings.dynamic import JSSPDynamicEmbedding
from rl4co.models.nn.graph.hgnn import HetGNNEncoder
from rl4co.models.nn.mlp import MLP
from rl4co.models.zoo.am.decoder import AttentionModelDecoder, PrecomputedCache
from rl4co.utils.ops import batchify, gather_by_index

from .encoder import GCN4JSSP


class L2DActor(nn.Module, metaclass=abc.ABCMeta):
    """Base decoder model for actor in L2D. The actor is responsible for generating the logits for the action
    similar to the decoder in autoregressive models. Since the decoder in L2D can have the additional purpose
    of extracting features (i.e. encoding the environment in ever iteration), we need an additional actor class.
    This function serves as template for such actor classes in L2D
    """

    @abc.abstractmethod
    def forward(
        self, td: TensorDict, hidden: Any = None, num_starts: int = 0
    ) -> Tuple[Tensor, Tensor]:
        """Obtain logits for current action to the next ones

        Args:
            td: TensorDict containing the input data
            hidden: Hidden state from the encoder. Can be any type
            num_starts: Number of starts for multistart decoding

        Returns:
            Tuple containing the logits and the action mask
        """
        raise NotImplementedError("Implement me in subclass!")

    def pre_decoder_hook(
        self, td: TensorDict, env=None, hidden: Any = None, num_starts: int = 0
    ) -> Tuple[TensorDict, Any]:
        """By default, we only require the input for the actor to be a tuple
        (in JSSP we only have operation embeddings but in FJSP we have operation
        and machine embeddings. By expecting a tuple we can generalize things.)

        Args:
            td: TensorDict containing the input data
            hidden: Hidden state from the encoder
            num_starts: Number of starts for multistart decoding

        Returns:
            Tuple containing the updated hidden state(s) and the input TensorDict
        """

        hidden = (hidden,) if not isinstance(hidden, tuple) else hidden

        if num_starts > 1:
            # NOTE: when using pomo, we need this
            hidden = tuple(map(lambda x: batchify(x, num_starts), hidden))

        return td, env, hidden


class JSSPActor(L2DActor):
    def __init__(
        self,
        embed_dim: int,
        hidden_dim: int,
        hidden_layers: int = 2,
        het_emb: bool = False,
        check_nan: bool = True,
    ) -> None:
        super().__init__()

        input_dim = (1 + int(het_emb)) * embed_dim
        self.mlp = MLP(
            input_dim=input_dim,
            output_dim=1,
            num_neurons=[hidden_dim] * hidden_layers,
            hidden_act="ReLU",
            out_act="Identity",
            input_norm="None",
            output_norm="None",
        )
        self.het_emb = het_emb
        self.dummy = nn.Parameter(torch.rand(input_dim))
        self.check_nan = check_nan

    def forward(self, td, op_emb, ma_emb=None):
        bs = td.size(0)
        # (bs, n_j)
        next_op = td["next_op"]
        # (bs, n_j, emb)
        job_emb = gather_by_index(op_emb, next_op, dim=1)
        if ma_emb is not None:
            ma_emb_per_op = einsum(td["ops_ma_adj"], ma_emb, "b m o, b m e -> b o e")
            # (bs, n_j, emb)
            ma_emb_per_job = gather_by_index(ma_emb_per_op, next_op, dim=1)
            # (bs, n_j, 2 * emb)
            job_emb = torch.cat((job_emb, ma_emb_per_job), dim=2)
        # (bs, n_j, 2 * emb)
        no_ops = self.dummy[None, None].expand(bs, 1, -1)
        # (bs, 1 + n_j, 2 * emb)
        all_actions = torch.cat((no_ops, job_emb), 1)
        # (bs, 1 + n_j)
        logits = self.mlp(all_actions).squeeze(2)

        if self.check_nan:
            assert not torch.isnan(logits).any(), "Logits contain NaNs"

        # (b, 1 + j)
        mask = td["action_mask"]

        return logits, mask


class FJSPActor(L2DActor):
    def __init__(
        self,
        embed_dim: int,
        hidden_dim: int,
        hidden_layers: int = 2,
        check_nan: bool = True,
    ) -> None:
        super().__init__()
        self.mlp = MLP(
            input_dim=2 * embed_dim,
            output_dim=1,
            num_neurons=[hidden_dim] * hidden_layers,
            hidden_act="ReLU",
            out_act="Identity",
            input_norm="None",
            output_norm="None",
        )
        self.dummy = nn.Parameter(torch.rand(2 * embed_dim))
        self.check_nan = check_nan

    def forward(self, td, ops_emb, ma_emb):
        bs, n_ma = ma_emb.shape[:2]
        # (bs, n_jobs, emb)
        job_emb = gather_by_index(ops_emb, td["next_op"], squeeze=False)
        # (bs, n_jobs, n_ma, emb)
        job_emb_expanded = job_emb.unsqueeze(2).expand(-1, -1, n_ma, -1)
        ma_emb_expanded = ma_emb.unsqueeze(1).expand_as(job_emb_expanded)
        # (bs, num_jobs * n_ma, 2*emb)
        h_actions = torch.cat((job_emb_expanded, ma_emb_expanded), dim=-1).flatten(1, 2)
        # (bs, 1, 2*emb_dim)
        no_ops = self.dummy[None, None].expand(bs, 1, -1)
        # (bs, num_jobs * n_ma + 1, 2*emb_dim)
        h_actions_w_noop = torch.cat((no_ops, h_actions), 1)
        # (b, j*m)
        logits = self.mlp(h_actions_w_noop).squeeze(-1)

        if self.check_nan:
            assert not torch.isnan(logits).any(), "Logits contain NaNs"
        # (b, 1 + j)
        mask = td["action_mask"]
        return logits, mask


class L2DDecoder(AutoregressiveDecoder):
    # feature extractor + actor
    def __init__(
        self,
        env_name: str = "jssp",
        feature_extractor: nn.Module = None,
        actor: nn.Module = None,
        init_embedding: nn.Module = None,
        embed_dim: int = 128,
        actor_hidden_dim: int = 128,
        actor_hidden_layers: int = 2,
        num_encoder_layers: int = 3,
        normalization: str = "batch",
        het_emb: bool = False,
        stepwise: bool = False,
        scaling_factor: int = 1000,
    ):
        super(L2DDecoder, self).__init__()

        if feature_extractor is None and stepwise:
            if env_name == "fjsp" or (het_emb and env_name == "jssp"):
                feature_extractor = HetGNNEncoder(
                    env_name=env_name,
                    embed_dim=embed_dim,
                    num_layers=num_encoder_layers,
                    normalization=normalization,
                    init_embedding=init_embedding,
                    scaling_factor=scaling_factor,
                )
            else:
                feature_extractor = GCN4JSSP(
                    embed_dim,
                    num_encoder_layers,
                    init_embedding=init_embedding,
                    scaling_factor=scaling_factor,
                )

        self.feature_extractor = feature_extractor

        if actor is None:
            if env_name == "fjsp":
                actor = FJSPActor(
                    embed_dim=embed_dim,
                    hidden_dim=actor_hidden_dim,
                    hidden_layers=actor_hidden_layers,
                )
            else:
                actor = JSSPActor(
                    embed_dim=embed_dim,
                    hidden_dim=actor_hidden_dim,
                    hidden_layers=actor_hidden_layers,
                    het_emb=het_emb,
                )

        self.actor = actor

    def forward(self, td, hidden, num_starts):
        if hidden is None:
            # NOTE in case we have multiple starts, td is batchified
            # (through decoding strategy pre decoding hook). Thus the
            # embeddings from feature_extractor have the correct shape
            num_starts = 0
            # (bs, n_j * n_ops, e), (bs, n_m, e)
            hidden, _ = self.feature_extractor(td)

        td, _, hidden = self.actor.pre_decoder_hook(td, None, hidden, num_starts)

        # (bs, n_j, e)
        logits, mask = self.actor(td, *hidden)

        return logits, mask


class L2DAttnPointer(PointerAttention):
    def __init__(
        self,
        env_name: str,
        embed_dim: int,
        num_heads: int,
        out_bias: bool = False,
        check_nan: bool = True,
    ):
        super().__init__(
            embed_dim=embed_dim,
            num_heads=num_heads,
            mask_inner=False,
            out_bias=out_bias,
            check_nan=check_nan,
        )
        self.env_name = env_name

    def forward(self, query, key, value, logit_key, attn_mask=None):
        # bs = query.size(0)
        # (b m j)
        logits = super().forward(query, key, value, logit_key, attn_mask=attn_mask)
        if self.env_name == "jssp":
            # (b j)
            logits = logits.sum(1)
        elif self.env_name == "fjsp":
            no_op_logits = logits[..., 0].sum(1, keepdims=True)
            logits = rearrange(logits[..., 1:], "b m j -> b (j m)")
            logits = torch.cat((no_op_logits, logits), dim=1)

        return logits


class AttnActor(AttentionModelDecoder):
    def __init__(
        self,
        embed_dim: int = 128,
        num_heads: int = 8,
        env_name: str = "tsp",
        context_embedding: nn.Module = None,
        dynamic_embedding: nn.Module = None,
        mask_inner: bool = True,
        out_bias_pointer_attn: bool = False,
        linear_bias: bool = False,
        use_graph_context: bool = True,
        check_nan: bool = True,
        sdpa_fn: callable = None,
        pointer: nn.Module = None,
        moe_kwargs: dict = None,
    ):
        super().__init__(
            embed_dim,
            num_heads,
            env_name,
            context_embedding,
            dynamic_embedding,
            mask_inner,
            out_bias_pointer_attn,
            linear_bias,
            use_graph_context,
            check_nan,
            sdpa_fn,
            pointer,
            moe_kwargs,
        )

    def pre_decoder_hook(
        self, td: TensorDict, env=None, hidden: Any = None, num_starts: int = 0
    ) -> Tuple[TensorDict, Any]:
        cache = self._precompute_cache(hidden, num_starts=num_starts)
        return td, env, (cache,)


class L2DAttnActor(AttnActor):
    def __init__(
        self,
        embed_dim: int = 128,
        num_heads: int = 8,
        env_name: str = "jssp",
        scaling_factor: int = 1000,
        stepwise: bool = False,
    ):
        context_embedding = SchedulingContext(embed_dim, scaling_factor=scaling_factor)
        if stepwise:
            # in a stepwise encoding setting, the embeddings contain all current information
            dynamic_embedding = None
        else:
            # otherwise we might want to update the static embeddings using dynamic updates
            dynamic_embedding = JSSPDynamicEmbedding(
                embed_dim, scaling_factor=scaling_factor
            )
        pointer = L2DAttnPointer(env_name, embed_dim, num_heads, check_nan=False)

        super().__init__(
            embed_dim=embed_dim,
            num_heads=num_heads,
            env_name=env_name,
            context_embedding=context_embedding,
            dynamic_embedding=dynamic_embedding,
            pointer=pointer,
        )
        self.dummy = nn.Parameter(torch.rand(1, embed_dim))

    def _compute_q(self, cached: PrecomputedCache, td: TensorDict):
        embeddings = cached.node_embeddings
        ma_embs = embeddings["machine_embeddings"]
        return self.context_embedding(ma_embs, td)

    def _compute_kvl(self, cached: PrecomputedCache, td: TensorDict):
        glimpse_k_stat, glimpse_v_stat, logit_k_stat = (
            gather_by_index(cached.glimpse_key, td["next_op"], dim=1),
            gather_by_index(cached.glimpse_val, td["next_op"], dim=1),
            gather_by_index(cached.logit_key, td["next_op"], dim=1),
        )
        # Compute dynamic embeddings and add to static embeddings
        glimpse_k_dyn, glimpse_v_dyn, logit_k_dyn = self.dynamic_embedding(td, cached)
        glimpse_k = glimpse_k_stat + glimpse_k_dyn
        glimpse_v = glimpse_v_stat + glimpse_v_dyn
        logit_k = logit_k_stat + logit_k_dyn

        no_ops = self.dummy.unsqueeze(1).expand(td.size(0), 1, -1).to(logit_k)
        logit_k = torch.cat((no_ops, logit_k), dim=1)

        return glimpse_k, glimpse_v, logit_k

    def _precompute_cache(self, embeddings: Tuple[torch.Tensor, torch.Tensor], **kwargs):
        ops_emb, ma_emb = embeddings

        (
            glimpse_key_fixed,
            glimpse_val_fixed,
            logit_key,
        ) = self.project_node_embeddings(
            ops_emb
        ).chunk(3, dim=-1)

        embeddings = TensorDict(
            {"op_embeddings": ops_emb, "machine_embeddings": ma_emb},
            batch_size=ops_emb.size(0),
        )
        # Organize in a dataclass for easy access
        return PrecomputedCache(
            node_embeddings=embeddings,
            graph_context=0,
            glimpse_key=glimpse_key_fixed,
            glimpse_val=glimpse_val_fixed,
            logit_key=logit_key,
        )



import torch.nn as nn

from rl4co.models.nn.env_embeddings import env_init_embedding
from rl4co.models.nn.graph.attnnet import Normalization, SkipConnection
from rl4co.models.zoo.ham.attention import HeterogenousMHA


class HeterogeneuousMHALayer(nn.Sequential):
    def __init__(
        self,
        num_heads,
        embed_dim,
        feedforward_hidden=512,
        normalization="batch",
    ):
        super(HeterogeneuousMHALayer, self).__init__(
            SkipConnection(HeterogenousMHA(num_heads, embed_dim, embed_dim)),
            Normalization(embed_dim, normalization),
            SkipConnection(
                nn.Sequential(
                    nn.Linear(embed_dim, feedforward_hidden),
                    nn.ReLU(),
                    nn.Linear(feedforward_hidden, embed_dim),
                )
                if feedforward_hidden > 0
                else nn.Linear(embed_dim, embed_dim)
            ),
            Normalization(embed_dim, normalization),
        )


class GraphHeterogeneousAttentionEncoder(nn.Module):
    def __init__(
        self,
        init_embedding=None,
        num_heads=8,
        embed_dim=128,
        num_encoder_layers=3,
        env_name=None,
        normalization="batch",
        feedforward_hidden=512,
        sdpa_fn=None,
    ):
        super(GraphHeterogeneousAttentionEncoder, self).__init__()

        # substitute env_name with pdp if none
        if env_name is None:
            env_name = "pdp"
        # Map input to embedding space
        if init_embedding is None:
            self.init_embedding = env_init_embedding(env_name, {"embed_dim": embed_dim})
        else:
            self.init_embedding = init_embedding

        self.layers = nn.Sequential(
            *(
                HeterogeneuousMHALayer(
                    num_heads,
                    embed_dim,
                    feedforward_hidden,
                    normalization,
                )
                for _ in range(num_encoder_layers)
            )
        )

    def forward(self, x, mask=None):
        assert mask is None, "Mask not yet supported!"
        # initial Embedding from features
        init_embeds = self.init_embedding(x)  # (batch_size, graph_size, embed_dim)
        # layers  (batch_size, graph_size, embed_dim)
        embeds = self.layers(init_embeds)
        return embeds, init_embeds



from typing import Callable

import torch

from tensordict.tensordict import TensorDict
from torch.distributions import Uniform

from rl4co.envs.routing.cvrp.generator import CVRPGenerator
from rl4co.utils.ops import get_distance


class CVRPTWGenerator(CVRPGenerator):
    """Data generator for the Capacitated Vehicle Routing Problem with Time Windows (CVRPTW) environment
    Generates time windows and service durations for the locations. The depot has a time window of [0, self.max_time].
    The time windows define the time span within which a service has to be started. To reach the depot in time from the last node,
    the end time of each node is bounded by the service duration and the distance back to the depot.
    The start times of the time windows are bounded by how long it takes to travel there from the depot.

    Args:
        num_loc: number of locations (customers) in the VRP, without the depot. (e.g. 10 means 10 locs + 1 depot)
        min_loc: minimum value for the location coordinates
        max_loc: maximum value for the location coordinates, default is 150 insted of 1.0, will be scaled
        loc_distribution: distribution for the location coordinates
        depot_distribution: distribution for the depot location. If None, sample the depot from the locations
        min_demand: minimum value for the demand of each customer
        max_demand: maximum value for the demand of each customer
        demand_distribution: distribution for the demand of each customer
        capacity: capacity of the vehicle
        max_time: maximum time for the vehicle to complete the tour
        scale: if True, the locations, time windows, and service durations will be scaled to [0, 1]. Default to False

    Returns:
        A TensorDict with the following keys:
            locs [batch_size, num_loc, 2]: locations of each city
            depot [batch_size, 2]: location of the depot
            demand [batch_size, num_loc]: demand of each customer
                while the demand of the depot is a placeholder
            capacity [batch_size, 1]: capacity of the vehicle
            durations [batch_size, num_loc]: service durations of each location
            time_windows [batch_size, num_loc, 2]: time windows of each location
    """

    def __init__(
        self,
        num_loc: int = 20,
        min_loc: float = 0.0,
        max_loc: float = 150.0,
        loc_distribution: int | float | str | type | Callable = Uniform,
        depot_distribution: int | float | str | type | Callable = Uniform,
        min_demand: int = 1,
        max_demand: int = 10,
        demand_distribution: int | float | str | type | Callable = Uniform,
        vehicle_capacity: float = 1.0,
        capacity: float = None,
        max_time: float = 480,
        scale: bool = False,
        **kwargs,
    ):
        super().__init__(
            num_loc=num_loc,
            min_loc=min_loc,
            max_loc=max_loc,
            loc_distribution=loc_distribution,
            depot_distribution=depot_distribution,
            min_demand=min_demand,
            max_demand=max_demand,
            demand_distribution=demand_distribution,
            vehicle_capacity=vehicle_capacity,
            capacity=capacity,
            **kwargs,
        )
        self.max_loc = max_loc
        self.min_time = 0.0
        self.max_time = max_time
        self.scale = scale

    def _generate(self, batch_size) -> TensorDict:
        td = super()._generate(batch_size)

        batch_size = [batch_size] if isinstance(batch_size, int) else batch_size

        ## define service durations
        # generate randomly (first assume service durations of 0, to be changed later)
        durations = torch.zeros(*batch_size, self.num_loc + 1, dtype=torch.float32)

        ## define time windows
        # 1. get distances from depot
        dist = get_distance(td["depot"], td["locs"].transpose(0, 1)).transpose(0, 1)
        dist = torch.cat((torch.zeros(*batch_size, 1), dist), dim=1)

        # 2. define upper bound for time windows to make sure the vehicle can get back to the depot in time
        upper_bound = self.max_time - dist - durations

        # 3. create random values between 0 and 1
        ts_1 = torch.rand(*batch_size, self.num_loc + 1)
        ts_2 = torch.rand(*batch_size, self.num_loc + 1)

        # 4. scale values to lie between their respective min_time and max_time and convert to integer values
        min_ts = (dist + (upper_bound - dist) * ts_1).int()
        max_ts = (dist + (upper_bound - dist) * ts_2).int()

        # 5. set the lower value to min, the higher to max
        min_times = torch.min(min_ts, max_ts)
        max_times = torch.max(min_ts, max_ts)

        # 6. reset times for depot
        min_times[..., :, 0] = 0.0
        max_times[..., :, 0] = self.max_time

        # 7. ensure min_times < max_times to prevent numerical errors in attention.py
        # min_times == max_times may lead to nan values in _inner_mha()
        mask = min_times == max_times
        if torch.any(mask):
            min_tmp = min_times.clone()
            min_tmp[mask] = torch.max(
                dist[mask].int(), min_tmp[mask] - 1
            )  # we are handling integer values, so we can simply substract 1
            min_times = min_tmp

            mask = min_times == max_times  # update mask to new min_times
            if torch.any(mask):
                max_tmp = max_times.clone()
                max_tmp[mask] = torch.min(
                    torch.floor(upper_bound[mask]).int(),
                    torch.max(
                        torch.ceil(min_tmp[mask] + durations[mask]).int(),
                        max_tmp[mask] + 1,
                    ),
                )
                max_times = max_tmp

        # Scale to [0, 1]
        if self.scale:
            durations = durations / self.max_time
            min_times = min_times / self.max_time
            max_times = max_times / self.max_time
            td["depot"] = td["depot"] / self.max_time
            td["locs"] = td["locs"] / self.max_time

        # 8. stack to tensor time_windows
        time_windows = torch.stack((min_times, max_times), dim=-1)

        assert torch.all(
            min_times < max_times
        ), "Please make sure the relation between max_loc and max_time allows for feasible solutions."

        # Reset duration at depot to 0
        durations[:, 0] = 0.0
        td.update(
            {
                "durations": durations,
                "time_windows": time_windows,
            }
        )
        return td



from typing import Any, Callable

import torch.nn as nn

from rl4co.data.transforms import StateAugmentation
from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl.reinforce.reinforce import REINFORCE
from rl4co.models.zoo.am import AttentionModelPolicy
from rl4co.utils.ops import gather_by_index, unbatchify
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class POMO(REINFORCE):
    """POMO Model for neural combinatorial optimization based on REINFORCE
    Based on Kwon et al. (2020) http://arxiv.org/abs/2010.16011.

    Note:
        If no policy kwargs is passed, we use the Attention Model policy with the following arguments:
        Differently to the base class:
        - `num_encoder_layers=6` (instead of 3)
        - `normalization="instance"` (instead of "batch")
        - `use_graph_context=False` (instead of True)
        The latter is due to the fact that the paper does not use the graph context in the policy, which seems to be
        helpful in overfitting to the training graph size.

    Args:
        env: TorchRL Environment
        policy: Policy to use for the algorithm
        policy_kwargs: Keyword arguments for policy
        baseline: Baseline to use for the algorithm. Note that POMO only supports shared baseline,
            so we will throw an error if anything else is passed.
        num_augment: Number of augmentations (used only for validation and test)
        augment_fn: Function to use for augmentation, defaulting to dihedral8
        first_aug_identity: Whether to include the identity augmentation in the first position
        feats: List of features to augment
        num_starts: Number of starts for multi-start. If None, use the number of available actions
        **kwargs: Keyword arguments passed to the superclass
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: nn.Module = None,
        policy_kwargs={},
        baseline: str = "shared",
        num_augment: int = 8,
        augment_fn: str | Callable = "dihedral8",
        first_aug_identity: bool = True,
        feats: list = None,
        num_starts: int = None,
        **kwargs,
    ):
        self.save_hyperparameters(logger=False)

        if policy is None:
            policy_kwargs_with_defaults = {
                "num_encoder_layers": 6,
                "normalization": "instance",
                "use_graph_context": False,
            }
            policy_kwargs_with_defaults.update(policy_kwargs)
            policy = AttentionModelPolicy(
                env_name=env.name, **policy_kwargs_with_defaults
            )

        assert baseline == "shared", "POMO only supports shared baseline"

        # Initialize with the shared baseline
        super(POMO, self).__init__(env, policy, baseline, **kwargs)

        self.num_starts = num_starts
        self.num_augment = num_augment
        if self.num_augment > 1:
            self.augment = StateAugmentation(
                num_augment=self.num_augment,
                augment_fn=augment_fn,
                first_aug_identity=first_aug_identity,
                feats=feats,
            )
        else:
            self.augment = None

        # Add `_multistart` to decode type for train, val and test in policy
        for phase in ["train", "val", "test"]:
            self.set_decode_type_multistart(phase)

    def shared_step(
        self, batch: Any, batch_idx: int, phase: str, dataloader_idx: int = None
    ):
        td = self.env.reset(batch)
        n_aug, n_start = self.num_augment, self.num_starts
        n_start = self.env.get_num_starts(td) if n_start is None else n_start

        # During training, we do not augment the data
        if phase == "train":
            n_aug = 0
        elif n_aug > 1:
            td = self.augment(td)

        # Evaluate policy
        out = self.policy(td, self.env, phase=phase, num_starts=n_start)

        # Unbatchify reward to [batch_size, num_augment, num_starts].
        reward = unbatchify(out["reward"], (n_aug, n_start))

        # Training phase
        if phase == "train":
            assert n_start > 1, "num_starts must be > 1 during training"
            log_likelihood = unbatchify(out["log_likelihood"], (n_aug, n_start))
            self.calculate_loss(td, batch, out, reward, log_likelihood)
            max_reward, max_idxs = reward.max(dim=-1)
            out.update({"max_reward": max_reward})
        # Get multi-start (=POMO) rewards and best actions only during validation and test
        else:
            if n_start > 1:
                # max multi-start reward
                max_reward, max_idxs = reward.max(dim=-1)
                out.update({"max_reward": max_reward})

                if out.get("actions", None) is not None:
                    # Reshape batch to [batch_size, num_augment, num_starts, ...]
                    actions = unbatchify(out["actions"], (n_aug, n_start))
                    out.update(
                        {
                            "best_multistart_actions": gather_by_index(
                                actions, max_idxs, dim=max_idxs.dim()
                            )
                        }
                    )
                    out["actions"] = actions

            # Get augmentation score only during inference
            if n_aug > 1:
                # If multistart is enabled, we use the best multistart rewards
                reward_ = max_reward if n_start > 1 else reward
                max_aug_reward, max_idxs = reward_.max(dim=1)
                out.update({"max_aug_reward": max_aug_reward})

                if out.get("actions", None) is not None:
                    actions_ = (
                        out["best_multistart_actions"] if n_start > 1 else out["actions"]
                    )
                    out.update({"best_aug_actions": gather_by_index(actions_, max_idxs)})

        metrics = self.log_metrics(out, phase, dataloader_idx=dataloader_idx)
        return {"loss": out.get("loss", None), **metrics}



import abc

from rl4co.models.common.constructive.base import ConstructiveDecoder


class AutoregressiveDecoder(ConstructiveDecoder, metaclass=abc.ABCMeta):
    """Template class for an autoregressive decoder, simple wrapper around
    :class:`rl4co.models.common.constructive.base.ConstructiveDecoder`

    Tip:
        This class will not work as it is and is just a template.
        An example for autoregressive encoder can be found as :class:`rl4co.models.zoo.am.decoder.AttentionModelDecoder`.
    """



from typing import Any, Generator, NamedTuple, Optional

import numba as nb
import numpy as np
import torch

from tensordict import TensorDict


class SubTSPMapping(NamedTuple):
    map_action_index: np.ndarray
    map_node_index: np.ndarray
    subprob_coordinates: torch.Tensor


class VRP2SubTSPAdapter:
    """Adapter class for decomposing and composing CVRP solutions.
    The library numba is used to accelerate the computation.
    """

    subproblem_env_name = "tsp"

    def __init__(
        self,
        parent_td: TensorDict,
        actions: torch.Tensor,
        subprob_batch_size: Optional[int] = None,
        min_node_count: int = 4,
    ) -> None:
        batch_size = parent_td.batch_size[0]
        n_samples = actions.shape[0] // batch_size
        assert actions.shape[0] == n_samples * batch_size

        # prepend depot to each route
        self._actions = np.concatenate(
            [
                np.zeros((actions.shape[0], 1), dtype=np.int64),
                actions.cpu().numpy(),
            ],
            axis=1,
        )

        self.map_action_index = _cvrp_action_partitioner(self._actions, min_node_count)
        self.coordinates = parent_td["locs"].cpu().numpy()
        self.subprob_batch_size = subprob_batch_size

    def get_actions(self):
        return torch.from_numpy(self._actions)

    def get_batched_subprobs(self) -> Generator[SubTSPMapping, Any, None]:
        batch_size = self.subprob_batch_size
        if batch_size is None:  # fallback to single batch
            map_node_index, subtsp_coordinates = _compose_subtsp_coordinates(
                self._actions, self.map_action_index, self.coordinates
            )
            yield SubTSPMapping(
                self.map_action_index,
                map_node_index,
                torch.from_numpy(subtsp_coordinates),
            )
        else:
            # group sub problems by node count
            node_count = self.map_action_index[:, 2] - self.map_action_index[:, 1]
            order = np.argsort(node_count)
            if len(order) == 0:
                return
            for start_index in range(0, len(order), batch_size):
                selected_subtsp_index = order[start_index : start_index + batch_size]
                map_action_index = self.map_action_index[selected_subtsp_index]
                map_node_index, subtsp_coordinates = _compose_subtsp_coordinates(
                    self._actions, map_action_index, self.coordinates
                )
                yield SubTSPMapping(
                    map_action_index,
                    map_node_index,
                    torch.from_numpy(subtsp_coordinates),
                )

    def update_actions(self, mapping: SubTSPMapping, subtsp_actions: torch.Tensor):
        subtsp_actions_np = subtsp_actions.cpu().numpy()
        assert subtsp_actions.shape == mapping.map_node_index.shape
        _update_cvrp_actions(
            self._actions,
            subtsp_actions_np,
            mapping.map_action_index,
            mapping.map_node_index,
        )


@nb.njit(nogil=True, parallel=False, cache=True)
def _cvrp_action_partitioner(routes: np.ndarray, min_node_count: int = 4):
    map_action_index = []
    route_length = routes.shape[1]
    for index in range(routes.shape[0]):
        start = 0
        last_is_not_zero = routes[index, 0] != 0
        for idx in range(1, route_length):
            node = routes[index, idx]
            if node == 0:
                if last_is_not_zero and idx - start >= min_node_count:
                    map_action_index.append((index, start, idx))
                last_is_not_zero = False
                start = idx
            else:
                last_is_not_zero = True
        if node != 0 and route_length - start >= min_node_count:  # handle final routes
            map_action_index.append((index, start, route_length))
    map_action_index = np.array(map_action_index, dtype=np.int32)
    return map_action_index


@nb.njit(nogil=True, parallel=True, cache=True)
def _compose_subtsp_coordinates(
    actions: np.ndarray, map_action_index: np.ndarray, coordinates: np.ndarray
):
    n_subtsp = map_action_index.shape[0]
    batch_size = coordinates.shape[0]
    n_samples = actions.shape[0] // batch_size
    max_subtsp_length = (map_action_index[:, 2] - map_action_index[:, 1]).max()
    subtsp_index = np.zeros((n_subtsp, max_subtsp_length + 1), dtype=np.int32)
    subtsp_coordinates = np.zeros(
        (n_subtsp, max_subtsp_length + 1, 2), dtype=coordinates.dtype
    )
    for idx in nb.prange(n_subtsp):
        route_idx, start, end = map_action_index[idx]
        inst_idx = route_idx // n_samples
        subtsp_index[idx, : end - start] = actions[route_idx, start:end]
        subtsp_coordinates[idx, :, :] = coordinates[inst_idx, subtsp_index[idx]]
    return subtsp_index, subtsp_coordinates


@nb.njit(nogil=True, parallel=True, cache=True)
def _update_cvrp_actions(
    cvrp_actions: np.ndarray,
    subtsp_actions: np.ndarray,
    map_action_index: np.ndarray,
    map_node_index: np.ndarray,
):
    subtsp_length = subtsp_actions.shape[1]
    subtsp_underlying_actions = np.take_along_axis(
        map_node_index, subtsp_actions, axis=-1
    )
    for idx in nb.prange(subtsp_actions.shape[0]):
        route_idx, start, end = map_action_index[idx]
        real_nodes = subtsp_underlying_actions[idx]

        # roll the route to ensure it starts with non-zero nodes
        # e.g. [1,2,0,0,3,4] -> [3,4,1,2,0,0]
        shift_count = 0
        last_node = real_nodes[-1]
        for _node_idx in range(subtsp_length):
            current_node = real_nodes[_node_idx]
            if current_node != 0 and last_node == 0:
                shift_count = _node_idx
                break
            last_node = current_node
        if shift_count != 0:
            real_nodes = np.roll(real_nodes, -shift_count)

        # update cvrp_actions
        cvrp_actions[route_idx, start + 1 : end] = real_nodes[real_nodes != 0]



import copy

import torch.nn as nn

from rl4co.envs import RL4COEnvBase
from rl4co.models.rl import PPO
from rl4co.models.rl.common.critic import CriticNetwork
from rl4co.models.zoo.am.policy import AttentionModelPolicy
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class AMPPO(PPO):
    """PPO Model based on Proximal Policy Optimization (PPO) with an attention model policy.
    We default to the attention model policy and the Attention Critic Network.

    Args:
        env: Environment to use for the algorithm
        policy: Policy to use for the algorithm
        critic: Critic to use for the algorithm
        policy_kwargs: Keyword arguments for policy
        critic_kwargs: Keyword arguments for critic
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: nn.Module = None,
        critic: CriticNetwork = None,
        policy_kwargs: dict = {},
        critic_kwargs: dict = {},
        **kwargs,
    ):
        if policy is None:
            policy = AttentionModelPolicy(env_name=env.name, **policy_kwargs)

        if critic is None:
            log.info("Creating critic network for {}".format(env.name))
            # we reuse the parameters of the model
            encoder = getattr(policy, "encoder", None)
            if encoder is None:
                raise ValueError("Critic network requires an encoder")
            critic = CriticNetwork(
                copy.deepcopy(encoder).to(next(encoder.parameters()).device),
                **critic_kwargs,
            )

        super().__init__(env, policy, critic, **kwargs)



import logging

from typing import Any, Callable, Optional

import torch

from tensordict import TensorDict

from rl4co.data.transforms import StateAugmentation
from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl.reinforce.reinforce import REINFORCE
from rl4co.models.zoo.polynet.policy import PolyNetPolicy
from rl4co.utils.ops import gather_by_index, unbatchify
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class PolyNet(REINFORCE):
    """PolyNet
    Based on Hottung et al. (2024) https://arxiv.org/abs/2402.14048.

    Note:
        PolyNet allows to learn diverse solution stratgies with a single model. This is achieved
        through a modified decoder and the Poppy loss (Grinsztajn et al. (2021)). PolyNet can be used with the attention model encoder or the MatNet encoder by
        setting encoder_type to "AM" or "MatNet", respectively.

    Args:
        env: TorchRL Environment
        policy: Policy to use for the algorithm
        k: Number of strategies to learn ("K" in the paper)
        val_num_solutions: Number of solutions that are generated per instance during validation
        encoder_type: Type of encoder that should be used. "AM" or "MatNet" are supported
        policy_kwargs: Keyword arguments for policy
        baseline: Baseline to use for the algorithm. Note that PolyNet only supports shared baseline,
            so we will throw an error if anything else is passed.
        num_augment: Number of augmentations (used only for validation and test)
        augment_fn: Function to use for augmentation, defaulting to dihedral8
        first_aug_identity: Whether to include the identity augmentation in the first position
        feats: List of features to augment
        **kwargs: Keyword arguments passed to the superclass
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: PolyNetPolicy = None,
        k: int = 128,
        val_num_solutions: int = 800,
        encoder_type="AM",
        base_model_checkpoint_path: str = None,
        policy_kwargs={},
        baseline: str = "shared",
        num_augment: int = 8,
        augment_fn: str | Callable = "dihedral8",
        first_aug_identity: bool = True,
        feats: list = None,
        **kwargs,
    ):
        self.save_hyperparameters(logger=False)

        self.k = k
        self.val_num_solutions = val_num_solutions

        assert encoder_type in [
            "AM",
            "MatNet",
        ], "Supported encoder types are 'AM' and 'MatNet'"

        assert baseline == "shared", "PolyNet only supports shared baseline"

        if (
            policy_kwargs.get("val_decode_type") == "greedy"
            or policy_kwargs.get("test_decode_type") == "greedy"
        ):
            assert (
                val_num_solutions <= k
            ), "If greedy decoding is used val_num_solutions must be <= k"

        if encoder_type == "MatNet":
            assert (
                num_augment == 1
            ), "MatNet does not use symmetric or dihedral augmentation"

        if policy is None:
            policy = PolyNetPolicy(
                env_name=env.name, k=k, encoder_type=encoder_type, **policy_kwargs
            )

        if base_model_checkpoint_path is not None:
            logging.info(
                f"Trying to load weights from baseline model {base_model_checkpoint_path}"
            )
            checkpoint = torch.load(base_model_checkpoint_path, weights_only=False)
            state_dict = checkpoint["state_dict"]
            state_dict = {k.replace("policy.", "", 1): v for k, v in state_dict.items()}
            policy.load_state_dict(state_dict, strict=False)

        train_batch_size = kwargs["batch_size"] if "batch_size" in kwargs else 64
        kwargs_with_defaults = {
            "val_batch_size": train_batch_size,
            "test_batch_size": train_batch_size,
        }
        kwargs_with_defaults.update(kwargs)

        # Initialize with the shared baseline
        super(PolyNet, self).__init__(env, policy, baseline, **kwargs_with_defaults)

        self.num_augment = num_augment
        if self.num_augment > 1:
            self.augment = StateAugmentation(
                num_augment=self.num_augment,
                augment_fn=augment_fn,
                first_aug_identity=first_aug_identity,
                feats=feats,
            )
        else:
            self.augment = None

        # Add `_multistart` to decode type for train, val and test in policy
        # for phase in ["train", "val", "test"]:
        #    self.set_decode_type_multistart(phase)

    def shared_step(
        self, batch: Any, batch_idx: int, phase: str, dataloader_idx: int = None
    ):
        td = self.env.reset(batch)
        n_aug = self.num_augment

        # During training, we do not augment the data
        if phase == "train":
            n_aug = 0
        elif n_aug > 1:
            td = self.augment(td)

        if phase == "train":
            n_start = self.k
        else:
            n_start = self.val_num_solutions

        # Evaluate policy
        out = self.policy(
            td,
            self.env,
            phase=phase,
            num_starts=n_start,
            multisample=True,
        )

        # Unbatchify reward to [batch_size, num_augment, num_starts].
        reward = unbatchify(out["reward"], (n_aug, n_start))

        # Training phase
        if phase == "train":
            assert n_start > 1, "num_starts must be > 1 during training"
            log_likelihood = unbatchify(out["log_likelihood"], (n_aug, n_start))
            self.calculate_loss(td, batch, out, reward, log_likelihood)
            max_reward, max_idxs = reward.max(dim=-1)
            out.update({"max_reward": max_reward})
        # Get multi-start (=POMO) rewards and best actions only during validation and test
        else:
            if n_start > 1:
                # max multi-start reward
                max_reward, max_idxs = reward.max(dim=-1)
                out.update({"max_reward": max_reward})

                if out.get("actions", None) is not None:
                    # Reshape batch to [batch_size, num_augment, num_starts, ...]
                    actions = unbatchify(out["actions"], (n_aug, n_start))
                    out.update(
                        {
                            "best_multistart_actions": gather_by_index(
                                actions, max_idxs.unsqueeze(2), dim=2
                            )
                        }
                    )
                    out["actions"] = actions

            # Get augmentation score only during inference
            if n_aug > 1:
                # If multistart is enabled, we use the best multistart rewards
                reward_ = max_reward if n_start > 1 else reward
                max_aug_reward, max_idxs = reward_.max(dim=1)
                out.update({"max_aug_reward": max_aug_reward})

                if out.get("actions", None) is not None:
                    actions_ = (
                        out["best_multistart_actions"] if n_start > 1 else out["actions"]
                    )
                    out.update({"best_aug_actions": gather_by_index(actions_, max_idxs)})

        metrics = self.log_metrics(out, phase, dataloader_idx=dataloader_idx)
        return {"loss": out.get("loss", None), **metrics}

    def calculate_loss(
        self,
        td: TensorDict,
        batch: TensorDict,
        policy_out: dict,
        reward: Optional[torch.Tensor] = None,
        log_likelihood: Optional[torch.Tensor] = None,
    ):
        """Calculate loss following Poppy (https://arxiv.org/abs/2210.03475).

        Args:
            td: TensorDict containing the current state of the environment
            batch: Batch of data. This is used to get the extra loss terms, e.g., REINFORCE baseline
            policy_out: Output of the policy network
            reward: Reward tensor. If None, it is taken from `policy_out`
            log_likelihood: Log-likelihood tensor. If None, it is taken from `policy_out`
        """
        # Extra: this is used for additional loss terms, e.g., REINFORCE baseline
        extra = batch.get("extra", None)
        reward = reward if reward is not None else policy_out["reward"]
        log_likelihood = (
            log_likelihood if log_likelihood is not None else policy_out["log_likelihood"]
        )

        # REINFORCE baseline
        bl_val, bl_loss = (
            self.baseline.eval(td, reward, self.env) if extra is None else (extra, 0)
        )

        # Log-likelihood mask. Mask everything but the best rollout per instance
        best_idx = (-reward).argsort(1).argsort(1)
        mask = best_idx < 1

        # Main loss function
        advantage = reward - bl_val  # advantage = reward - baseline
        reinforce_loss = -(advantage * log_likelihood * mask).mean()
        loss = reinforce_loss + bl_loss
        policy_out.update(
            {
                "loss": loss,
                "reinforce_loss": reinforce_loss,
                "bl_loss": bl_loss,
                "bl_val": bl_val,
            }
        )
        return policy_out



from typing import Any, Optional, Union

from tensordict import TensorDict
import torch

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl import REINFORCE
from rl4co.models.rl.reinforce.baselines import REINFORCEBaseline
from rl4co.models.zoo.deepaco.policy import DeepACOPolicy


class DeepACO(REINFORCE):
    """Implements DeepACO: https://arxiv.org/abs/2309.14032.

    Args:
        env: Environment to use for the algorithm
        policy: Policy to use for the algorithm
        baseline: REINFORCE baseline. Defaults to "no" because the shared baseline is manually implemented.
        train_with_local_search: Whether to train with local search. Defaults to False.
        ls_reward_aug_W: Coefficient to be used for the reward augmentation with the local search. Defaults to 0.95.
        policy_kwargs: Keyword arguments for policy
        baseline_kwargs: Keyword arguments for baseline
        **kwargs: Keyword arguments passed to the superclass
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: Optional[DeepACOPolicy] = None,
        baseline: Union[REINFORCEBaseline, str] = "no",  # Shared baseline is manually implemented
        train_with_local_search: bool = True,
        ls_reward_aug_W: float = 0.95,
        policy_kwargs: dict = {},
        baseline_kwargs: dict = {},
        **kwargs,
    ):
        if policy is None:
            policy = DeepACOPolicy(
                env_name=env.name, train_with_local_search=train_with_local_search, **policy_kwargs
            )

        super().__init__(env, policy, baseline, baseline_kwargs, **kwargs)

        self.train_with_local_search = train_with_local_search
        self.ls_reward_aug_W = ls_reward_aug_W

    def shared_step(
        self, batch: Any, batch_idx: int, phase: str, dataloader_idx: Optional[int] = None
    ):
        td = self.env.reset(batch)
        # Perform forward pass (i.e., constructing solution and computing log-likelihoods)
        out = self.policy(td, self.env, phase=phase)

        # Compute loss
        if phase == "train":
            out["loss"] = self.calculate_loss(td, batch, out)

        metrics = self.log_metrics(out, phase, dataloader_idx=dataloader_idx)
        return {"loss": out.get("loss", None), **metrics}

    def calculate_loss(
        self,
        td: TensorDict,
        batch: TensorDict,
        policy_out: dict,
        reward: Optional[torch.Tensor] = None,
        log_likelihood: Optional[torch.Tensor] = None,
    ):
        """Calculate loss for REINFORCE algorithm.

        Args:
            td: TensorDict containing the current state of the environment
            batch: Batch of data. This is used to get the extra loss terms, e.g., REINFORCE baseline
            policy_out: Output of the policy network
            reward: Reward tensor. If None, it is taken from `policy_out`
            log_likelihood: Log-likelihood tensor. If None, it is taken from `policy_out`
        """
        reward = policy_out["reward"]
        advantage = reward - reward.mean(dim=1, keepdim=True)  # Shared baseline

        if self.train_with_local_search:
            ls_reward = policy_out["ls_reward"]
            ls_advantage = ls_reward - ls_reward.mean(dim=1, keepdim=True)  # Shared baseline
            weighted_advantage = advantage * (1 - self.ls_reward_aug_W) + ls_advantage * self.ls_reward_aug_W
        else:
            weighted_advantage = advantage

        return -(weighted_advantage * policy_out["log_likelihood"]).mean()



import abc

from typing import Callable

import torch

from tensordict.tensordict import TensorDict
from torch.distributions import Exponential, Normal, Poisson, Uniform

from rl4co.envs.common.distribution_utils import (
    Cluster,
    Gaussian_Mixture,
    Mix_Distribution,
    Mix_Multi_Distributions,
    Mixed,
)


class Generator(metaclass=abc.ABCMeta):
    """Base data generator class, to be called with `env.generator(batch_size)`"""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, batch_size) -> TensorDict:
        batch_size = [batch_size] if isinstance(batch_size, int) else batch_size
        return self._generate(batch_size)

    @abc.abstractmethod
    def _generate(self, batch_size, **kwargs) -> TensorDict:
        raise NotImplementedError


def get_sampler(
    val_name: str,
    distribution: int | float | str | type | Callable,
    low: float = 0,
    high: float = 1.0,
    **kwargs,
):
    """Get the sampler for the variable with the given distribution.
    If kwargs are passed, they will be parsed e.g. with `val_name` + `_dist_arg` (e.g. `loc_std` for Normal distribution).

    Args:
        val_name: Name of the variable
        distribution: int/float value (as constant distribution), or string with the distribution name (supporting
            uniform, normal, exponential, and poisson) or PyTorch Distribution type or a callable function that
            returns a PyTorch Distribution
        low: Minimum value for the variable, used for Uniform distribution
        high: Maximum value for the variable, used for Uniform distribution
        kwargs: Additional arguments for the distribution

    Example:
        ```python
        sampler_uniform = get_sampler("loc", "uniform", 0, 1)
        sampler_normal = get_sampler("loc", "normal", loc_mean=0.5, loc_std=.2)
        ```
    """
    if isinstance(distribution, (int, float)):
        return Uniform(low=distribution, high=distribution)
    elif distribution == Uniform or distribution == "uniform":
        return Uniform(low=low, high=high)
    elif distribution == Normal or distribution == "normal" or distribution == "gaussian":
        assert (
            kwargs.get(val_name + "_mean", None) is not None
        ), "mean is required for Normal distribution"
        assert (
            kwargs.get(val_name + "_std", None) is not None
        ), "std is required for Normal distribution"
        return Normal(loc=kwargs[val_name + "_mean"], scale=kwargs[val_name + "_std"])
    elif distribution == Exponential or distribution == "exponential":
        assert (
            kwargs.get(val_name + "_rate", None) is not None
        ), "rate is required for Exponential/Poisson distribution"
        return Exponential(rate=kwargs[val_name + "_rate"])
    elif distribution == Poisson or distribution == "poisson":
        assert (
            kwargs.get(val_name + "_rate", None) is not None
        ), "rate is required for Exponential/Poisson distribution"
        return Poisson(rate=kwargs[val_name + "_rate"])
    elif distribution == "center":
        return Uniform(low=(high - low) / 2, high=(high - low) / 2)
    elif distribution == "corner":
        return Uniform(
            low=low, high=low
        )  # todo: should be also `low, high` and any other corner
    elif isinstance(distribution, Callable):
        return distribution(**kwargs)
    elif distribution == "gaussian_mixture":
        return Gaussian_Mixture(num_modes=kwargs["num_modes"], cdist=kwargs["cdist"])
    elif distribution == "cluster":
        return Cluster(kwargs["n_cluster"])
    elif distribution == "mixed":
        return Mixed(kwargs["n_cluster_mix"])
    elif distribution == "mix_distribution":
        return Mix_Distribution(kwargs["n_cluster"], kwargs["n_cluster_mix"])
    elif distribution == "mix_multi_distributions":
        return Mix_Multi_Distributions()
    else:
        raise ValueError(f"Invalid distribution type of {distribution}")


def batch_to_scalar(param):
    """Return first element if in batch. Used for batched parameters that are the same for all elements in the batch."""
    if len(param.shape) > 0:
        return param[0].item()
    if isinstance(param, torch.Tensor):
        return param.item()
    return param



from rl4co.models.common.constructive.nonautoregressive.decoder import (
    NonAutoregressiveDecoder,
)
from rl4co.models.common.constructive.nonautoregressive.encoder import (
    NonAutoregressiveEncoder,
)
from rl4co.models.common.constructive.nonautoregressive.policy import (
    NonAutoregressivePolicy,
)



from pathlib import Path
from typing import Tuple

import torch

from tensordict import TensorDict

ProcessingData = list[Tuple[int, int]]


def parse_job_line(line: Tuple[int]) -> Tuple[ProcessingData]:
    """
    Parses a JSSP job data line of the following form:

        <num operations> * (<machine> <processing time>)

    In words, a line consist of n_ops pairs of values, where the first value is the
    machine identifier and the second value is the processing time of the corresponding
    operation-machine combination

    Note that the machine indices start from 1, so we subtract 1 to make them
    zero-based.
    """

    operations = []
    i = 0

    while i < len(line):
        machine = int(line[i])
        duration = int(line[i + 1])
        operations.append((machine, duration))
        i += 2

    return operations


def get_n_ops_of_instance(file):
    lines = file2lines(file)
    jobs = [parse_job_line(line) for line in lines[1:]]
    n_ope_per_job = torch.Tensor([len(x) for x in jobs]).unsqueeze(0)
    total_ops = int(n_ope_per_job.sum())
    return total_ops


def get_max_ops_from_files(files):
    return max(map(get_n_ops_of_instance, files))


def read(loc: Path, max_ops=None):
    """
    Reads an JSSP instance.

    Args:
        loc: location of instance file
        max_ops: optionally specify the maximum number of total operations (will be filled by padding)

    Returns:
        instance: the parsed instance
    """
    lines = file2lines(loc)

    # First line contains metadata.
    num_jobs, num_machines = lines[0][0], lines[0][1]

    # The remaining lines contain the job-operation data, where each line
    # represents a job and its operations.
    jobs = [parse_job_line(line) for line in lines[1:]]
    n_ope_per_job = torch.Tensor([len(x) for x in jobs]).unsqueeze(0)
    total_ops = int(n_ope_per_job.sum())
    if max_ops is not None:
        assert total_ops <= max_ops, "got more operations then specified through max_ops"
    max_ops = max_ops or total_ops
    max_ops_per_job = int(n_ope_per_job.max())

    end_op_per_job = n_ope_per_job.cumsum(1) - 1
    start_op_per_job = torch.cat((torch.zeros((1, 1)), end_op_per_job[:, :-1] + 1), dim=1)

    pad_mask = torch.arange(max_ops)
    pad_mask = pad_mask.ge(total_ops).unsqueeze(0)

    proc_times = torch.zeros((num_machines, max_ops))
    op_cnt = 0
    for job in jobs:
        for ma, dur in job:
            # subtract one to let indices start from zero
            proc_times[ma - 1, op_cnt] = dur
            op_cnt += 1
    proc_times = proc_times.unsqueeze(0)

    td = TensorDict(
        {
            "start_op_per_job": start_op_per_job,
            "end_op_per_job": end_op_per_job,
            "proc_times": proc_times,
            "pad_mask": pad_mask,
        },
        batch_size=[1],
    )

    return td, num_jobs, num_machines, max_ops_per_job


def file2lines(loc: Path | str) -> list[list[int]]:
    with open(loc, "r") as fh:
        lines = [line for line in fh.readlines() if line.strip()]

    def parse_num(word: str):
        return int(word) if "." not in word else int(float(word))

    return [[parse_num(x) for x in line.split()] for line in lines]



import hydra

from lightning import Callback
from lightning.pytorch.loggers import Logger
from omegaconf import DictConfig

from rl4co.utils import pylogger

log = pylogger.get_pylogger(__name__)


def instantiate_callbacks(callbacks_cfg: DictConfig) -> list[Callback]:
    """Instantiates callbacks from config."""

    callbacks: list[Callback] = []

    if not callbacks_cfg:
        log.warning("No callback configs found! Skipping..")
        return callbacks

    if not isinstance(callbacks_cfg, DictConfig):
        raise TypeError("Callbacks config must be a DictConfig!")

    for _, cb_conf in callbacks_cfg.items():
        if isinstance(cb_conf, DictConfig) and "_target_" in cb_conf:
            log.info(f"Instantiating callback <{cb_conf._target_}>")
            callbacks.append(hydra.utils.instantiate(cb_conf))

    return callbacks


def instantiate_loggers(logger_cfg: DictConfig, model) -> list[Logger]:
    """Instantiates loggers from config."""

    logger_list: list[Logger] = []

    if not logger_cfg:
        log.warning("No logger configs found! Skipping...")
        return logger_list

    if not isinstance(logger_cfg, DictConfig):
        raise TypeError("Logger config must be a DictConfig!")

    for _, lg_conf in logger_cfg.items():
        if isinstance(lg_conf, DictConfig) and "_target_" in lg_conf:
            log.info(f"Instantiating logger <{lg_conf._target_}>")
            if hasattr(lg_conf, "log_gradients"):
                log_gradients = lg_conf.get("log_gradients", False)
                # manually remove parameter, since pop doesnt work on DictConfig
                del lg_conf.log_gradients
            else:
                log_gradients = False
            logger = hydra.utils.instantiate(lg_conf)
            if hasattr(logger, "watch") and log_gradients:
                # make use of wandb gradient statistics logger
                logger.watch(model, log_graph=False)
            logger_list.append(logger)

    return logger_list



from typing import Optional
from tensordict import TensorDict
import torch.nn as nn

from rl4co.models.zoo.nargnn.encoder import NARGNNEncoder


class GFACSEncoder(NARGNNEncoder):
    """
    NARGNNEncoder with log-partition function estimation for training with
    Trajectory Balance (TB) loss (Malkin et al., https://arxiv.org/abs/2201.13259)
    """
    def __init__(
        self,
        embed_dim: int = 64,
        env_name: str = "tsp",
        # TODO: pass network
        init_embedding: Optional[nn.Module] = None,
        edge_embedding: Optional[nn.Module] = None,
        graph_network: Optional[nn.Module] = None,
        heatmap_generator: Optional[nn.Module] = None,
        num_layers_heatmap_generator: int = 5,
        num_layers_graph_encoder: int = 15,
        act_fn="silu",
        agg_fn="mean",
        linear_bias: bool = True,
        k_sparse: Optional[int] = None,
        z_out_dim: int = 1,
    ):
        super().__init__(
            embed_dim=embed_dim,
            env_name=env_name,
            init_embedding=init_embedding,
            edge_embedding=edge_embedding,
            graph_network=graph_network,
            heatmap_generator=heatmap_generator,
            num_layers_heatmap_generator=num_layers_heatmap_generator,
            num_layers_graph_encoder=num_layers_graph_encoder,
            act_fn=act_fn,
            agg_fn=agg_fn,
            linear_bias=linear_bias,
            k_sparse=k_sparse,
        )
        self.Z_net = nn.Sequential(
            nn.Linear(embed_dim, embed_dim), nn.SiLU(), nn.Linear(embed_dim, z_out_dim)
        )
        self.z_out_dim = z_out_dim

    def forward(self, td: TensorDict):
        """Forward pass of the encoder.
        Transform the input TensorDict into the latent representation.
        """
        # Transfer to embedding space
        node_embed = self.init_embedding(td)
        graph = self.edge_embedding(td, node_embed)

        # Process embedding into graph
        # TODO: standardize?
        graph.x, graph.edge_attr = self.graph_network(
            graph.x, graph.edge_index, graph.edge_attr
        )

        logZ = self.Z_net(graph.edge_attr).reshape(-1, len(td), self.z_out_dim).mean(0)

        # Generate heatmap logits
        heatmap_logits = self.heatmap_generator(graph)

        # Return latent representation (i.e. heatmap logits), initial embeddings and logZ
        return heatmap_logits, node_embed, logZ



import torch

from tensordict.tensordict import TensorDict

from rl4co.envs.common.utils import Generator
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class FFSPGenerator(Generator):
    """Data generator for the Flow Shop Scheduling Problem (FFSP).

    Args:
        num_stage: number of stages
        num_machine: number of machines
        num_job: number of jobs
        min_time: minimum running time of each job on each machine
        max_time: maximum running time of each job on each machine
        flatten_stages: whether to flatten the stages

    Returns:
        A TensorDict with the following key:
            run_time [batch_size, num_job, num_machine, num_stage]: running time of each job on each machine

    Note:
        - [IMPORTANT] This version of ffsp requires the number of machines in each stage to be the same
    """

    def __init__(
        self,
        num_stage: int = 2,
        num_machine: int = 3,
        num_job: int = 4,
        min_time: int = 2,
        max_time: int = 10,
        flatten_stages: bool = True,
        **unused_kwargs,
    ):
        self.num_stage = num_stage
        self.num_machine = num_machine
        self.num_machine_total = num_machine * num_stage
        self.num_job = num_job
        self.min_time = min_time
        self.max_time = max_time
        self.flatten_stages = flatten_stages

        # FFSP environment doen't have any other kwargs
        if len(unused_kwargs) > 0:
            log.error(f"Found {len(unused_kwargs)} unused kwargs: {unused_kwargs}")

    def _generate(self, batch_size) -> TensorDict:
        # Init observation: running time of each job on each machine
        run_time = torch.randint(
            low=self.min_time,
            high=self.max_time,
            size=(*batch_size, self.num_job, self.num_machine_total),
        )

        return TensorDict(
            {
                "run_time": run_time,
            },
            batch_size=batch_size,
        )
